import os

from tests.base import TestBase

from aim.sdk import Run
from aim.storage.context import Context
from aim.storage.containertreeview import ContainerTreeView
from aim.storage.rockscontainer import RocksContainer
from aim.storage.hashing import hash_auto


class TestRunContainerData(TestBase):
    def test_meta_tree_contexts_and_names(self):
        train_context = {'subset': 'train'}
        val_context = {'subset': 'val'}
        empty_context = {}

        run = Run(repo=self.repo, system_tracking_interval=None)
        run.track(1.0, name='metric 1', context=train_context)
        run.track(1.0, name='metric 2', context=train_context)
        run.track(1.0, name='metric 1', context=val_context)
        run.track(1.0, name='metric 2', context=val_context)
        run.track(1.0, name='metric 3', context=val_context)
        run.track(0.0, name='metric')

        meta_container_path = os.path.join(self.repo.path, 'meta', 'chunks', run.hash)
        rc = RocksContainer(meta_container_path, read_only=True)
        tree = ContainerTreeView(rc)

        contexts = tree[('meta', 'contexts')]
        for ctx in [train_context, val_context, empty_context]:
            self.assertIn(Context(ctx).idx, contexts)
            self.assertDictEqual(contexts[Context(ctx).idx], ctx)

        self.assertSetEqual({'float'}, set(tree.view(('meta', 'traces_types')).keys()))
        metric_traces = tree.view(('meta', 'traces_types', 'float', Context(train_context).idx)).collect()
        self.assertSetEqual({'metric 1', 'metric 2'}, set(metric_traces.keys()))
        metric_traces = tree.view(('meta', 'traces_types', 'float', Context(val_context).idx)).collect()
        self.assertSetEqual({'metric 1', 'metric 2', 'metric 3'}, set(metric_traces.keys()))
        metric_traces = tree.view(('meta', 'traces_types', 'float', Context(empty_context).idx)).collect()
        self.assertSetEqual({'metric'}, set(metric_traces.keys()))

    def test_meta_run_tree_contexts_and_names(self):
        train_context = {'subset': 'train'}
        val_context = {'subset': 'val'}
        empty_context = {}

        run = Run(repo=self.repo, system_tracking_interval=None)
        run.track(1, name='metric 1', context=train_context)
        run.track(1, name='metric 2', context=train_context)
        run.track(1, name='metric 1', context=val_context)
        run.track(1, name='metric 2', context=val_context)
        run.track(1, name='metric 3', context=val_context)
        run.track(0, name='metric')

        meta_container_path = os.path.join(self.repo.path, 'meta', 'chunks', run.hash)
        rc = RocksContainer(meta_container_path, read_only=True)
        tree = ContainerTreeView(rc)

        contexts = tree.view(('meta', 'chunks', run.hash, 'contexts')).collect()
        for ctx in [train_context, val_context, empty_context]:
            self.assertIn(Context(ctx).idx, contexts)
            self.assertDictEqual(contexts[Context(ctx).idx], ctx)

        traces = tree.view(('meta', 'chunks', run.hash, 'traces', Context(train_context).idx)).collect()
        self.assertSetEqual({'metric 1', 'metric 2'}, set(traces.keys()))
        traces = tree.view(('meta', 'chunks', run.hash, 'traces', Context(val_context).idx)).collect()
        self.assertSetEqual({'metric 1', 'metric 2', 'metric 3'}, set(traces.keys()))
        traces = tree.view(('meta', 'chunks', run.hash, 'traces', Context(empty_context).idx)).collect()
        self.assertSetEqual({'metric'}, set(traces.keys()))

    def test_run_trace_dtype_and_last_value(self):
        run = Run()
        run.track(1.0, name='metric 1', context={})
        run.track(2.0, name='metric 1', context={})
        run.track(3.0, name='metric 1', context={})
        run.track(1.0, name='metric 1', context={'subset': 'train'})

        meta_container_path = os.path.join(self.repo.path, 'meta', 'chunks', run.hash)
        rc = RocksContainer(meta_container_path, read_only=True)
        tree = ContainerTreeView(rc)
        metric_1_dict = tree.view(('meta', 'chunks', run.hash, 'traces', Context({}).idx, 'metric 1')).collect()
        self.assertEqual(3.0, metric_1_dict['last'])
        self.assertEqual('float', metric_1_dict['dtype'])

        metric_1_dict = tree.view(('meta', 'chunks', run.hash, 'traces',
                                   Context({'subset': 'train'}).idx, 'metric 1')).collect()
        self.assertEqual(1.0, metric_1_dict['last'])

    def test_series_tree_values(self):
        # sequential steps
        run = Run()
        run.track(1.0, name='metric 1', context={})
        run.track(2.0, name='metric 1', context={})
        run.track(3.0, name='metric 1', context={})

        series_container_path = os.path.join(self.repo.path, 'seqs', 'chunks', run.hash)
        rc = RocksContainer(series_container_path, read_only=True)
        tree = ContainerTreeView(rc)
        trace = tree.view(('seqs', 'v2', 'chunks', run.hash, Context({}).idx, 'metric 1'))
        self.assertSetEqual({'step', 'val', 'epoch', 'time'}, set(trace.keys()))
        steps = sorted(trace.array('step').values_list())
        vals = trace.array('val').values_list()
        epochs = trace.array('epoch').values_list()
        times = trace.array('time').values_list()

        self.assertEqual(3, len(steps))
        self.assertEqual(3, len(vals))
        self.assertEqual(3, len(epochs))
        self.assertEqual(3, len(times))

        self.assertListEqual([0, 1, 2], steps)
        self.assertEqual(1.0, trace.array('val')[hash_auto(steps[0])])
        self.assertEqual(2.0, trace.array('val')[hash_auto(steps[1])])
        self.assertEqual(3.0, trace.array('val')[hash_auto(steps[2])])

        # user-specified steps
        run = Run()
        run.track(1.0, name='metric 1', step=10, context={})
        run.track(2.0, name='metric 1', step=20, context={})
        run.track(3.0, name='metric 1', step=30, context={})

        series_container_path = os.path.join(self.repo.path, 'seqs', 'chunks', run.hash)
        rc = RocksContainer(series_container_path, read_only=True)
        tree = ContainerTreeView(rc)
        trace = tree.view(('seqs', 'v2', 'chunks', run.hash, Context({}).idx, 'metric 1'))
        steps = sorted(trace.array('step').values_list())
        vals = trace.array('val').values_list()
        epochs = trace.array('epoch').values_list()
        times = trace.array('time').values_list()

        self.assertEqual(3, len(steps))
        self.assertEqual(3, len(vals))
        self.assertEqual(3, len(epochs))
        self.assertEqual(3, len(times))

        self.assertListEqual([10, 20, 30], steps)
        self.assertEqual(1.0, trace.array('val')[hash_auto(steps[0])])
        self.assertEqual(2.0, trace.array('val')[hash_auto(steps[1])])
        self.assertEqual(3.0, trace.array('val')[hash_auto(steps[2])])

        # user-specified steps, unordered
        run = Run()
        run.track(3.0, name='metric 1', step=30, context={})
        run.track(1.0, name='metric 1', step=10, context={})
        run.track(2.0, name='metric 1', step=20, context={})

        series_container_path = os.path.join(self.repo.path, 'seqs', 'chunks', run.hash)
        rc = RocksContainer(series_container_path, read_only=True)
        tree = ContainerTreeView(rc)
        trace = tree.view(('seqs', 'v2', 'chunks', run.hash, Context({}).idx, 'metric 1'))
        steps = sorted(trace.array('step').values_list())
        vals = trace.array('val').values_list()
        epochs = trace.array('epoch').values_list()
        times = trace.array('time').values_list()

        self.assertEqual(3, len(steps))
        self.assertEqual(3, len(vals))
        self.assertEqual(3, len(epochs))
        self.assertEqual(3, len(times))

        self.assertListEqual([10, 20, 30], steps)
        self.assertEqual(1.0, trace.array('val')[hash_auto(steps[0])])
        self.assertEqual(2.0, trace.array('val')[hash_auto(steps[1])])
        self.assertEqual(3.0, trace.array('val')[hash_auto(steps[2])])

    def test_run_set_param_meta_tree(self):
        run = Run()
        run['p1'] = 1
        run['p2'] = {'foo': 'bar'}
        run['p3'] = 0.0
        run['p4'] = True
        run['p5'] = ['x', 1, 2]
        run['p6'] = None
        run['p7'] = 'bazե\x00a-a'
        run['p8'] = b'blob'

        meta_container_path = os.path.join(self.repo.path, 'meta', 'chunks', run.hash)
        rc = RocksContainer(meta_container_path, read_only=True)
        tree = ContainerTreeView(rc)
        meta_attrs_tree = tree.view(('meta', 'attrs'))
        meta_run_attrs_tree = tree.view(('meta', 'chunks', run.hash, 'attrs'))

        val = meta_attrs_tree['p1']
        self.assertEqual(int, type(val))
        self.assertEqual(1, val)

        val = meta_run_attrs_tree['p1']
        self.assertEqual(int, type(val))
        self.assertEqual(1, val)

        val = meta_attrs_tree['p3']
        self.assertEqual(float, type(val))
        self.assertEqual(0.0, val)

        val = meta_run_attrs_tree['p3']
        self.assertEqual(float, type(val))
        self.assertEqual(0.0, val)

        val = meta_attrs_tree['p4']
        self.assertEqual(bool, type(val))
        self.assertEqual(True, val)

        val = meta_run_attrs_tree['p4']
        self.assertEqual(bool, type(val))
        self.assertEqual(True, val)

        val = meta_attrs_tree['p2']
        self.assertEqual(dict, type(val))
        self.assertEqual({'foo': 'bar'}, val)

        val = meta_run_attrs_tree['p2']
        self.assertEqual(dict, type(val))
        self.assertEqual({'foo': 'bar'}, val)

        val = meta_attrs_tree['p5']
        self.assertEqual(list, type(val))
        self.assertEqual(['x', 1, 2], val)

        val = meta_run_attrs_tree['p5']
        self.assertEqual(list, type(val))
        self.assertEqual(['x', 1, 2], val)

        val = meta_attrs_tree['p6']
        self.assertEqual(None, val)

        val = meta_run_attrs_tree['p6']
        self.assertEqual(None, val)

        val = meta_attrs_tree['p7']
        self.assertEqual(str, type(val))
        self.assertEqual('bazե\x00a-a', val)

        val = meta_run_attrs_tree['p7']
        self.assertEqual(str, type(val))
        self.assertEqual('bazե\x00a-a', val)

        val = meta_attrs_tree['p8']
        self.assertEqual(bytes, type(val))
        self.assertEqual(b'blob', val)

        val = meta_run_attrs_tree['p8']
        self.assertEqual(bytes, type(val))
        self.assertEqual(b'blob', val)



import shlex

from aim.cli.configs import VERSION_NAME, UP_NAME
import aim.cli.manager.manager as manager

# Error message prefix for aim commands
ERROR_MSG_PREFIX = b'Error:'

# returned by get_execution_context
_COLAB_EXEC_CONTEXT = "_COLAB_EXEC_CONTEXT"
_IPYTHON_EXEC_CONTEXT = "_IPYTHON_EXEC_CONTEXT"
_OTHER_EXEC_CONTEXT = "_OTHER_EXEC_CONTEXT"

# current execution context
_CURRENT_CONTEXT = _OTHER_EXEC_CONTEXT

# environment specific constants
# useful for detecting the environment from the UI
_SAGE_MAKER_NOTEBOOK_PATH_POSTFIX = "/aim-sage"
_NOTEBOOK_PATH_POSTFIX = "/notebook"


def get_execution_context():
    """Determine the most specific context that we're in.
    Returns:
      _COLAB_EXEC_CONTEXT: If in Colab with an IPython notebook context.
      _IPYTHON_EXEC_CONTEXT: If we are in an IPython notebook
        context but not in colab (i.e. `jupyter notebook`)
        line).
      _OTHER_EXEC_CONTEXT: Otherwise (e.g., by running a Python script at the
        command-line or using the `ipython` interactive shell).
    """
    # In Colab, the `google.colab` module is available, but the shell
    # returned by `IPython.get_ipython` does not have a `get_trait`
    # method.

    # imports are dynamic, since some modules are not available for all contexts
    try:
        import IPython
    except ImportError:
        pass
    else:
        ipython = IPython.get_ipython()
        # @TODO find a stable way to get colab context
        if ipython is not None and 'google.colab' in str(ipython):
            # We are in Colab notebook context
            # global _CURRENT_CONTEXT
            # _CURRENT_CONTEXT = _COLAB_EXEC_CONTEXT
            return _COLAB_EXEC_CONTEXT

        # In an IPython command line shell or Jupyter notebook
        elif ipython is not None and ipython.has_trait("kernel"):
            # global _CURRENT_CONTEXT
            # _CURRENT_CONTEXT = _IPYTHON_EXEC_CONTEXT
            return _IPYTHON_EXEC_CONTEXT

    # Otherwise, we're not in a known notebook context.
    return _OTHER_EXEC_CONTEXT


def get_argument_options(line):
    """
        Returns parsed argument options and command from magic cell as  dict (command, options)
        currently parse only --<name>=value style to dict
        Set default values for the required fields, otherwise the provided fields
        Will omit unsupported args @TODO notify about unsupported args
        @TODO add process args all styles to dict
    """
    # @TODO improve this logic
    # --proxy-url is useful to print the right url, and set UI's url into iframe correctly
    supported_args = ['--port', '--host', '--repo', '--proxy-url']

    args = shlex.split(line)
    command = args[0]

    options = {
        '--host': '127.0.0.1',
        '--port': '43801',
        '--base-path': _NOTEBOOK_PATH_POSTFIX
    }
    for arg in args[1:]:
        key, value = arg.split('=', 1)
        if key in supported_args:
            options[key] = value
    # if --proxy-url passed
    if options.get('--proxy-url'):
        options['--base-path'] = f'/proxy/absolute/{options["--port"]}{_SAGE_MAKER_NOTEBOOK_PATH_POSTFIX}'

    return command, options


def display_colab(port, display):
    """Display Aim instance in a Colab output frame.
       It need go through the proxy
    """
    import IPython.display

    shell = """
        (async () => {{
            const url = new URL('{path}/', await google.colab.kernel.proxyPort({port}, {{'cache': true}}));
            const iframe = document.createElement('iframe');
            iframe.src = url;
            const a = document.createElement('a');
            a.href = url;
            a.innerHTML = 'Open in new browser tab';
            a.setAttribute('target', '_blank');

            iframe.setAttribute('width', '100%');
            iframe.setAttribute('height', '800');
            iframe.setAttribute('frameborder', 0);
            document.body.appendChild(iframe);
            document.body.appendChild(a);
        }})();
        """.format(path=_NOTEBOOK_PATH_POSTFIX, port=port)

    script = IPython.display.Javascript(shell)

    if display:
        display.update(script)
    else:
        IPython.display.display(script)


def display_notebook(host, port, display, proxy_url=None):
    """Display Aim instance in an ipython context output frame.
    """
    import IPython.display
    url = "{}:{}{}".format(host, port, _NOTEBOOK_PATH_POSTFIX)

    # @TODO add warning if proxy_url is not defined
    if proxy_url:
        # jupyter-server-proxy supports absolute paths by using it with /proxy/absolute/<port> path
        url = "{}{}{}{}/".format(proxy_url, '/proxy/absolute/', port, _SAGE_MAKER_NOTEBOOK_PATH_POSTFIX)
        print(url)

    shell = """
              <iframe id="aim" width="100%" height="800" frameborder="0" src={}>
              </iframe>
            """.format(url)

    iframe = IPython.display.HTML(shell)
    display.update(iframe)


def up(options, context):
    """
        Calls to run `aim up` command width corresponding arguments
        Handles the result of the command
        Renders the <iframe> tag for the notebook and message for the shell users
        The <iframe> renders width the corresponding way for different execution contexts (mainly for notebooks)
    """

    try:
        import IPython
        import IPython.display
    except ImportError:
        IPython = None

    display = None
    if context == _OTHER_EXEC_CONTEXT:
        print("Launching Aim ...")
    else:
        display = IPython.display.display(
            IPython.display.Pretty("Launching Aim ..."),
            display_id=True,
        )

    result = manager.run_process(UP_NAME, options)

    if result.status == manager.ManagerActionStatuses.Failed:
        print(result.info["message"])
        return

    port = result.info["port"]
    host = result.info["host"]

    # successful exec of aim up command
    if context == _COLAB_EXEC_CONTEXT:
        display_colab(port, display)
        return
    if context == _IPYTHON_EXEC_CONTEXT:
        display_notebook(host, port, display, options.get("--proxy-url"))
        return

    # other context
    print("Open {}:{}".format(host, port))


def version(options, context):
    """Handles aim version (get version process) and send to the ui"""
    result = manager.run_process(VERSION_NAME, options)
    if result.status is manager.ManagerActionStatuses.Failed:
        print(result.info["message"])
    else:
        print("Aim v{}".format(result.info["version"]))


# Those are aim magic function available commands
# This is why we are not using constants from aim.cli.commands
# It is possible to add commands outside aim cli
handlers = {
    UP_NAME: up,
    VERSION_NAME: version
}


def execute_magic_aim(line):
    """ `aim` line magic function
        We are trying to keep similarity with the native aim cli commands as much as possible
    """
    context = get_execution_context()
    command, options = get_argument_options(line)
    # check command existence
    if command not in handlers:
        print('Invalid operation.')
        return

    # call corresponding handler
    handlers[command](options, context)


def load_ipython_extension(ipython):
    ipython.register_magic_function(execute_magic_aim, magic_kind="line", magic_name="aim")



from typing import Any, Iterator, Tuple, Union

from aim.storage import encoding
from aim.storage.encoding.encoding_native cimport decode_path

from aim.storage.types import AimObject, AimObjectPath

from aim.storage.types import BLOB
from aim.storage.utils import ArrayFlag, ObjectFlag, CustomObjectFlagType

from aim.storage.object import CustomObjectBase, CustomObject
from aim.storage.treeview import TreeView
from aim.storage.inmemorytreeview import InMemoryTreeView
from aim.storage.treeutils_non_native import convert_to_native_object


def unfold_tree(
    obj: AimObject,
    *,
    path: AimObjectPath = (),
    unfold_array: bool = True,
    depth: int = None,
    strict: bool = True,
) -> Iterator[Tuple[AimObjectPath, Any]]:
    if depth == 0:
        yield path, obj
        return
    if depth is not None:
        depth -= 1

    if obj is None:
        yield path, obj
    elif isinstance(obj, (bool, int, float, str, bytes)):
        yield path, obj
    elif isinstance(obj, BLOB):
        yield path, obj
    elif isinstance(obj, (list, tuple)):
        if not unfold_array:
            yield path, obj
        else:
            yield path, ArrayFlag
            # Ellipsis (...) is set when array elements are expected
            for idx, val in enumerate(obj):
                yield from unfold_tree(val, path=path + (idx,), unfold_array=unfold_array, depth=depth, strict=strict)
    elif isinstance(obj, dict):
        # TODO: set ObjectFlag for all dicts?
        if obj == {}:
            yield path, ObjectFlag
        for key, val in obj.items():
            yield from unfold_tree(val, path=path + (key,), unfold_array=unfold_array, depth=depth, strict=strict)
    elif isinstance(obj, CustomObjectBase):
        aim_name, aim_obj = obj._aim_encode()
        yield path, CustomObjectFlagType(aim_name)
        for key, val in obj.storage.items():
            yield from unfold_tree(val, path=path + (key,), unfold_array=unfold_array, depth=depth, strict=strict)
    elif isinstance(obj, TreeView):
        # TODO we need to implement TreeView.traverse()
        raise NotImplementedError
    else:
        obj = convert_to_native_object(obj, strict=strict)
        yield from unfold_tree(obj, path=path, unfold_array=unfold_array, depth=depth, strict=strict)



cpdef val_to_node(
    val: Any,
    strict: bool = True,
    resolve_objects: bool = False
):
    if not strict:
        node = dict()
        if val == ArrayFlag:
            node['__example_type__'] = str(list)
        elif val != ObjectFlag:
            node['__example_type__'] = str(type(val))
        return node
    if val == ObjectFlag:
        return dict()
    elif val == ArrayFlag:
        return []
    elif isinstance(val, CustomObjectFlagType):
        if resolve_objects:
            return dict()
        else:
            return CustomObject._aim_decode(val.aim_name, InMemoryTreeView(container={}, constructed=False))
    else:
        return val


def fold_tree(
    paths_vals: Iterator[Tuple[AimObjectPath, Any]],
    strict: bool = True,
    resolve_objects: bool = False
) -> AimObject:
    (keys, val), = iter_fold_tree(paths_vals,
                                  level=0, strict=strict, resolve_objects=resolve_objects)
    return val


def iter_fold_tree(
    paths_vals: Iterator[Tuple[AimObjectPath, Any]],
    level: int = 0,
    strict: bool = True,
    resolve_objects: bool = False,
    skip_top_level: bool = False
):
    cdef int idx
    stack = []
    path = []

    try:
        keys, val = next(paths_vals)
        if keys:
            raise StopIteration
        node = val_to_node(val, resolve_objects=resolve_objects)
        # TODO [AT]: fix this
        if skip_top_level and node == []:
            node = {}
        stack.append(node)
    except StopIteration:
        if level > 0:
            return
        else:
            raise KeyError

    for keys, val in paths_vals:
        idx = 0
        while idx < len(path):
            if keys[idx] != path[idx]:
                break
            idx += 1

        while idx < len(path):
            last_state = stack.pop()
            if len(stack) == level:
                if isinstance(last_state, CustomObject):
                    last_state.storage._constructed = True
                yield tuple(path), last_state
            path.pop()

        node = val_to_node(val, strict=strict, resolve_objects=resolve_objects)

        if len(keys) == len(path):
            stack.pop()
            path.pop()

        assert len(keys) == len(path) + 1
        key_to_add = keys[-1]
        path.append(key_to_add)
        assert stack

        if isinstance(stack[-1], list):
            assert isinstance(key_to_add, int)
            if key_to_add < 0:
                raise NotImplementedError
            elif key_to_add < len(stack[-1]):
                stack[-1][key_to_add] = node
            else:
                while len(stack[-1]) != key_to_add:
                    stack[-1].append(None)
                stack[-1].append(node)
        elif isinstance(stack[-1], dict):
            stack[-1][key_to_add] = node
        elif isinstance(stack[-1], CustomObject):
            stack[-1].storage[key_to_add] = node
        else:
            raise ValueError
        stack.append(node)

    if level < len(stack):
        val = stack[level]
        if isinstance(val, CustomObject):
            val.storage._constructed = True
        yield tuple(path[:level]), val

def encode_paths_vals(
    paths_vals: Iterator[Tuple[AimObjectPath, Any]]
) -> Iterator[Tuple[bytes, bytes]]:
    for path, val in paths_vals:
        path = encoding.encode_path(path)
        val = encoding.encode(val)
        yield path, val


cdef class DecodePathsVals(object):
    cdef paths_vals
    cdef current_path
    cdef to_yield
    cdef int num_yielded

    def __cinit__(self, paths_vals):
        self.paths_vals = paths_vals
        self.current_path = None
        self.to_yield = []
        self.num_yielded = 0

    def __iter__(self):
        return self

    def __next__(self):
        return self._next()

    cdef _next(self):
        cdef int idx
        if self.to_yield:
            val = self.to_yield[self.num_yielded]
            self.num_yielded += 1
            if self.num_yielded == len(self.to_yield):
                self.num_yielded = 0
                self.to_yield = []
            return val

        while True:
            encoded_path, encoded_val = next(self.paths_vals)
            path = decode_path(encoded_path)
            val = encoding.decode(encoded_val)

            if self.current_path is None:
                if path:
                    to_yield = (), ObjectFlag
                    self.to_yield.append(to_yield)
                self.current_path = []
                break
            elif self.current_path != path:
                break

        idx = 0
        new_path = []
        while idx < len(self.current_path):
            key = path[idx]
            if key != self.current_path[idx]:
                break
            new_path.append(key)
            idx += 1
        self.current_path = new_path

        while idx < len(path):
            self.current_path.append(path[idx])
            if idx < len(path):
                to_yield = tuple(self.current_path), ObjectFlag
                self.to_yield.append(to_yield)
            idx += 1

        to_yield = tuple(path), val
        self.to_yield.append(to_yield)


        val = self.to_yield[self.num_yielded]
        self.num_yielded += 1
        if self.num_yielded == len(self.to_yield):
            self.num_yielded = 0
            self.to_yield = []
        return val


def encode_tree(
    obj: AimObject,
    strict: bool = True
) -> Iterator[Tuple[bytes, bytes]]:
    return encode_paths_vals(
        unfold_tree(obj, strict=strict)
    )


def decode_tree(
    paths_vals: Iterator[Tuple[bytes, bytes]],
    strict: bool = True,
    resolve_objects: bool = False
) -> AimObject:
    return fold_tree(
        DecodePathsVals(paths_vals),
        strict=strict,
        resolve_objects=resolve_objects
    )


def iter_decode_tree(
    paths_vals: Iterator[Tuple[bytes, bytes]],
    level: int = 1,
    skip_top_level: bool = False
):
    return iter_fold_tree(
        DecodePathsVals(paths_vals),
        level=level,
        skip_top_level=skip_top_level
    )



from typing import Dict, List, Tuple, Union

from aim.storage.utils import BLOB  # noqa F401

NoneType = type(None)


AimObjectKey = Union[int, str]
AimObjectPath = Tuple[AimObjectKey, ...]
AimObjectPrimitive = Union[
    NoneType,
    bool,
    int,
    float,
    str,
    bytes
]
AimObjectArray = Union[List['AimObject'], Tuple['AimObject', ...]]
AimObjectDict = Dict[AimObjectKey, 'AimObject']
AimObject = Union[
    AimObjectPrimitive,
    AimObjectArray,
    AimObjectDict
]


class Singleton(type):
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args,
                                                                 **kwargs)
        return cls._instances[cls]


class SafeNone(metaclass=Singleton):
    def get(self, item):
        return self

    def __repr__(self):
        return 'None'

    def __call__(self, *args, **kwargs):
        return self

    def __getattr__(self, item):
        return self

    def __getitem__(self, item):
        return self

    def __bool__(self):
        return False

    def __eq__(self, other):
        return other is None or isinstance(other, SafeNone)

    def __iter__(self):
        return self

    def __next__(self):
        raise StopIteration


class CustomObjectBase:
    pass



import pytz

from tests.base import PrefilledDataApiTestBase, ApiTestBase
from tests.utils import generate_image_set

from parameterized import parameterized
import datetime

from aim.sdk.run import Run


class TestProjectApi(PrefilledDataApiTestBase):
    def test_project_activity_api(self):
        with self.repo.structured_db as db:
            db.create_experiment('My experiment')

        experiment_count = len(self.repo.structured_db.experiments())
        run_count = len(self.repo.structured_db.runs())
        client = self.client
        response = client.get('/api/projects/activity')
        self.assertEqual(200, response.status_code)
        data = response.json()
        today_gmt = datetime.datetime.now().astimezone(pytz.timezone('gmt')).strftime('%Y-%m-%dT%H:00:00')
        self.assertEqual(run_count, data['num_runs'])
        self.assertEqual(run_count, data['activity_map'][today_gmt])
        self.assertEqual(experiment_count, data['num_experiments'])  # count 'default' experiment

    def test_project_params_api(self):
        client = self.client
        response = client.get('/api/projects/params')
        self.assertEqual(200, response.status_code)
        data = response.json()
        self.assertIn('params', data)
        self.assertIn('metric', data)
        self.assertIn('images', data)
        self.assertEqual({}, data['images'])
        self.assertSetEqual({'accuracy', 'loss'}, set(data['metric']))
        self.assertIn('hparams', data['params'])
        self.assertIn('batch_size', data['params']['hparams'])
        self.assertIn('lr', data['params']['hparams'])
        self.assertIn('name', data['params'])
        self.assertIn('run_index', data['params'])
        self.assertIn('start_time', data['params'])


class TestProjectParamsWithImagesApi(ApiTestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        run1 = Run(system_tracking_interval=None)
        run1.track(1., name='metric1', context={'a': True})
        run1.track(generate_image_set(1), name='images1', context={'a': True})
        run1.track(generate_image_set(1), name='images1', context={'b': True})

        run2 = Run(system_tracking_interval=None)
        run2.track(1, name='metric2', context={'a': True})
        run2.track(generate_image_set(1)[0], name='images2', context={'b': True})

    @parameterized.expand([
        ({'sequence': ('metric', 'images')},),  # metrics only
        (None,)                                 # default
    ])
    def test_project_images_and_metric_info_api(self, qparams):
        client = self.client
        response = client.get('/api/projects/params', params=qparams)
        self.assertEqual(200, response.status_code)
        data = response.json()

        self.assertIn('metric', data)
        self.assertIn('images', data)

        self.assertTrue({'metric1', 'metric2'}.issubset(set(data['metric'].keys())))
        self.assertTrue({'images1', 'images2'}.issubset(set(data['images'].keys())))

        self.assertEqual(1, len(data['metric']['metric1']))
        self.assertDictEqual({'a': 1}, data['metric']['metric1'][0])

        self.assertEqual(2, len(data['images']['images1']))

        self.assertEqual(1, len(data['metric']['metric2']))
        self.assertDictEqual({'a': 1}, data['metric']['metric2'][0])

        self.assertEqual(1, len(data['images']['images2']))
        self.assertDictEqual({'b': 1}, data['images']['images2'][0])

    def test_project_images_info_only_api(self):
        client = self.client
        response = client.get('/api/projects/params', params={'sequence': 'images'})
        self.assertEqual(200, response.status_code)
        data = response.json()

        self.assertNotIn('metric', data)
        self.assertIn('images', data)

    def test_project_metric_info_only_api(self):
        client = self.client
        response = client.get('/api/projects/params', params={'sequence': 'metric'})
        self.assertEqual(200, response.status_code)
        data = response.json()

        self.assertIn('metric', data)
        self.assertNotIn('images', data)

    def test_invalid_sequence_type(self):
        client = self.client
        response = client.get('/api/projects/params', params={'sequence': 'non-existing-sequence'})
        self.assertEqual(400, response.status_code)



from aim.storage import encoding as E
from aim.storage.encoding.encoding import decode
from aim.storage.object import CustomObject
from aim.storage.types import AimObject, AimObjectKey, AimObjectPath
from aim.storage.utils import ArrayFlag, CustomObjectFlagType
from aim.storage.container import Container
from aim.storage import treeutils
from aim.storage.treearrayview import TreeArrayView

from typing import Any, Iterator, Tuple, Union, List

from aim.storage.treeview import TreeView


class ContainerTreeView(TreeView):
    def __init__(
        self,
        container: Container
    ) -> None:
        self.container = container

    def preload(
        self
    ):
        self.container.preload()

    def finalize(
        self,
        index: 'ContainerTreeView'
    ):
        self.container.finalize(index=index.container)

    def view(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        resolve: bool = False
    ):
        prefix = E.encode_path(path)

        container_view = self.container.view(prefix)
        tree_view = ContainerTreeView(container_view)
        # Okay, but let's decide if we want to initialize a CustomObject view
        if not resolve:
            return tree_view

        flag = decode(container_view.get(b'', default=b'\0'))
        if isinstance(flag, CustomObjectFlagType):
            return CustomObject._aim_decode(flag.aim_name, tree_view)

        return tree_view

    def make_array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ):
        prefix = E.encode_path(path)
        self.container[prefix] = E.encode(ArrayFlag)

    def collect(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        strict: bool = True,
        resolve_objects: bool = False
    ) -> AimObject:
        if path == Ellipsis:
            path = ()
        if isinstance(path, (int, str)):
            path = (path,)
        prefix = E.encode_path(path)
        it = self.container.view(prefix).items()
        try:
            return treeutils.decode_tree(it, strict=strict, resolve_objects=resolve_objects)
        except KeyError:
            raise KeyError('No key {} is present.'.format(path))

    def __delitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath]
    ):
        if path == Ellipsis:
            path = ()
        if not isinstance(path, (tuple, list)):
            path = (path,)
        encoded_path = E.encode_path(path)
        self.container.delete_range(encoded_path, encoded_path + b'\xff')

    def set(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        value: AimObject,
        strict: bool = True
    ):
        if path == Ellipsis:
            path = ()
        if not isinstance(path, (tuple, list)):
            path = (path,)

        batch = self.container.batch()
        encoded_path = E.encode_path(path)
        self.container.delete_range(encoded_path, encoded_path + b'\xff',
                                    store_batch=batch)
        for key, val in treeutils.encode_tree(value, strict=strict):
            self.container.set(encoded_path + key, val,
                               store_batch=batch)
        self.container.commit(batch)

    def keys_eager(
            self,
            path: Union[AimObjectKey, AimObjectPath] = (),
    ) -> List[Union[AimObjectPath, AimObjectKey]]:
        return list(self.subtree(path).keys())

    def keys(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = 0
    ) -> Iterator[Union[AimObjectPath, AimObjectKey]]:
        encoded_path = E.encode_path(path)
        walker = self.container.walk(encoded_path)
        path = None
        while True:
            try:
                if path is None:
                    path = next(walker)
                else:
                    path = walker.send(path)
            except StopIteration:
                return
            path = E.decode_path(path)
            path = path[:max(level, 1)]
            if level <= 0:
                yield path[0]
            else:
                yield path
            p = E.encode_path(path)
            assert p.endswith(b'\xfe')
            path = p[:-1] + b'\xff'

    def items_eager(
            self,
            path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> List[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        return list(self.subtree(path).items())

    def items(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Iterator[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        prefix = E.encode_path(path)
        it = self.container.view(prefix).items()
        for path, value in treeutils.iter_decode_tree(it, level=1, skip_top_level=True):
            key, = path
            yield key, value

    def iterlevel(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = 1
    ) -> Iterator[Tuple[
        AimObjectPath,
        AimObject
    ]]:
        prefix = E.encode_path(path)
        it = self.container.items(prefix)
        for path, value in treeutils.iter_decode_tree(it, level=level, skip_top_level=True):
            yield path, value

    def array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        dtype: Any = None
    ) -> TreeArrayView:
        return TreeArrayView(self.subtree(path), dtype=dtype)

    def first_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        if isinstance(path, (int, str)):
            path = (path,)
        prefix = E.encode_path(path)
        p = E.decode_path(self.container.view(prefix).next_key())
        return p[0]

    def last_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        if isinstance(path, (int, str)):
            path = (path,)
        prefix = E.encode_path(path)
        # assert prefix.endswith(b'\xfe')
        # prefix = prefix[:-1] + b'\xff'
        p = E.decode_path(self.container.view(prefix).prev_key())
        if not p:
            raise KeyError
        return p[0]



from aim.sdk import Repo
from aim.sdk.types import QueryReportMode
from aim.web.api.runs.utils import get_run_props
from performance_tests.utils import timing


@timing()
def collect_runs_data(query):
    repo = Repo.default_repo()
    runs = repo.query_runs(query, report_mode=QueryReportMode.DISABLED)
    runs_dict = {}
    for run_trace_collection in runs.iter_runs():
        run = run_trace_collection.run
        runs_dict[run.hash] = {
            'params': run[...],
            'traces': run.collect_sequence_info(sequence_types='metric'),
            'props': get_run_props(run)
        }


@timing()
def collect_metrics_data(query):
    repo = Repo.default_repo()
    runs_dict = {}
    runs = repo.query_metrics(query=query, report_mode=QueryReportMode.DISABLED)
    for run_trace_collection in runs.iter_runs():
        run = None
        traces_list = []
        for trace in run_trace_collection.iter():
            if not run:
                run = run_trace_collection.run
            iters, values = trace.values.sparse_numpy()
            traces_list.append({
                'name': trace.name,
                'context': trace.context.to_dict(),
                'values': values,
                'iters': iters,
                'epochs': trace.epochs.values_numpy(),
                'timestamps': trace.timestamps.values_numpy()
            })
        if run:
            runs_dict[run.hash] = {
                'traces': traces_list,
                'params': run[...],
                'props': get_run_props(run),
            }


@timing()
def query_runs(query):
    repo = Repo.default_repo()
    runs = list(repo.query_runs(query=query, report_mode=QueryReportMode.DISABLED).iter_runs())


@timing()
def query_metrics(query):
    repo = Repo.default_repo()
    metrics = list(repo.query_metrics(query=query, report_mode=QueryReportMode.DISABLED).iter())




from typing import Tuple, Union

from aim.storage.object import CustomObject
from aim.sdk.sequence import MediaSequenceBase
from aim.sdk.objects.text import Text


@CustomObject.alias('aim.log_line')
class LogLine(Text):
    AIM_NAME = 'aim.log_line'

    def __init__(self, line):
        if not isinstance(line, str):
            raise TypeError('`line` must be a string.')

        super().__init__(line)


class Logs(MediaSequenceBase):
    """Class representing series of LogLine objects."""

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        text_typename = LogLine.get_typename()
        return text_typename

    @classmethod
    def sequence_name(cls) -> str:
        return 'logs'



import uuid
import sqlalchemy as sa

from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declared_attr

from aim.web.api.db import Base
from aim.web.api.utils import datetime_now


class AppMixin(object):
    uuid = sa.Column(sa.Text, primary_key=True)

    created_at = sa.Column(sa.DateTime, default=datetime_now)
    updated_at = sa.Column(sa.DateTime, default=datetime_now, onupdate=datetime_now)
    is_archived = sa.Column(sa.Boolean, default=False)

    @declared_attr
    def dashboard_id(cls):
        return sa.Column('dashboard_id', sa.ForeignKey('dashboards.uuid'))

    @declared_attr
    def dashboard(cls):
        return relationship("Dashboard")


class ExploreState(AppMixin, Base):
    __tablename__ = 'explore_states'

    type = sa.Column(sa.Text, nullable=False)
    state = sa.Column(sa.Text)

    def __init__(self):
        self.uuid = str(uuid.uuid1())
        self.is_archived = False



from parameterized import parameterized
import random

from tests.base import ApiTestBase
from tests.utils import decode_encoded_tree_stream, generate_image_set

from aim.storage.treeutils import decode_tree
from aim.storage.context import Context
from aim.sdk.run import Run


class TestNoImagesRunQueryApi(ApiTestBase):
    def test_query_images_api_empty_result(self):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/', params={'q': query, 'report_progress': False})
        self.assertEqual(200, response.status_code)
        self.assertEqual(b'', response.content)


class RunImagesTestBase(ApiTestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        run = cls.create_run(repo=cls.repo)
        run['images_per_step'] = 16
        for step in range(100):
            images = generate_image_set(img_count=16, caption_prefix=f'Image {step}')
            run.track(images, name='random_images')
            run.track(random.random(), name='random_values')
        cls.run_hash = run.hash


class TestRunImagesSearchApi(RunImagesTestBase):
    def test_query_images_api_defaults(self):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/', params={'q': query, 'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        self.assertEqual([0, 100], run_data['ranges']['record_range_total'])
        self.assertEqual([0, 100], run_data['ranges']['record_range_used'])
        self.assertEqual([0, 16], run_data['ranges']['index_range_total'])
        self.assertEqual([0, 16], run_data['ranges']['index_range_used'])
        self.assertEqual(16, run_data['params']['images_per_step'])

        trace_data = run_data['traces'][0]
        self.assertEqual('random_images', trace_data['name'])
        self.assertEqual(51, len(trace_data['iters']))
        self.assertEqual(4, trace_data['iters'][2])
        self.assertEqual(51, len(trace_data['values']))

        img_list = trace_data['values'][2]
        self.assertEqual(6, len(img_list))

        img = img_list[3]
        self.assertEqual('Image 4 9', img['caption'])
        self.assertEqual(9, img['index'])
        self.assertEqual(16, img['width'])
        self.assertEqual(16, img['height'])
        self.assertEqual('png', img['format'])
        self.assertIn('blob_uri', img)

    def test_query_images_api_custom_densities_dense(self):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/',
                              params={'q': query, 'record_density': 200, 'index_density': 10, 'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024),
                                                                  concat_chunks=True))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        trace_data = run_data['traces'][0]
        self.assertEqual(100, len(trace_data['values']))
        self.assertEqual(100, len(trace_data['iters']))
        self.assertEqual(16, len(trace_data['values'][0]))
        self.assertEqual(16, len(trace_data['values'][99]))

    def test_query_images_api_custom_densities_sparse(self):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/', params={'q': query,
                                                                  'record_density': 10,
                                                                  'index_density': 4,
                                                                  'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        trace_data = run_data['traces'][0]
        self.assertEqual(11, len(trace_data['values']))
        self.assertEqual(11, len(trace_data['iters']))
        self.assertEqual(4, len(trace_data['values'][0]))
        self.assertEqual(4, len(trace_data['values'][9]))
        image_indices = [img['index'] for img in trace_data['values'][0]]
        self.assertListEqual([0, 4, 8, 12], image_indices)
        image_names = [img['caption'] for img in trace_data['values'][9]]
        self.assertListEqual(['Image 90 0', 'Image 90 4', 'Image 90 8', 'Image 90 12'], image_names)

    @parameterized.expand([
        ('10:20', [0, 100], [10, 20], 10),
        (':30', [0, 100], [0, 30], 30),
        ('30:', [0, 100], [30, 100], 70)
    ])
    def test_query_images_api_custom_record_ranges(self, input_range, total_range, used_range, count):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/', params={'q': query,
                                                                  'record_range': input_range,
                                                                  'record_density': 100,
                                                                  'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        self.assertEqual(total_range, run_data['ranges']['record_range_total'])
        self.assertEqual(used_range, run_data['ranges']['record_range_used'])
        trace_data = run_data['traces'][0]
        self.assertEqual(count, len(trace_data['values']))
        self.assertEqual(count, len(trace_data['iters']))
        self.assertEqual(used_range[0], trace_data['iters'][0])
        self.assertEqual(used_range[-1] - 1, trace_data['iters'][-1])

    def test_query_images_api_calculate_ranges(self):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/', params={
            'q': query,
            'record_range': '10:20',
            'index_range': '3:6',
            'calc_ranges': True,
            'report_progress': False
        })
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        self.assertEqual([0, 100], run_data['ranges']['record_range_total'])
        self.assertEqual([10, 20], run_data['ranges']['record_range_used'])
        self.assertEqual([0, 16], run_data['ranges']['index_range_total'])
        self.assertEqual([3, 6], run_data['ranges']['index_range_used'])

        trace_data = run_data['traces'][0]
        self.assertEqual(10, len(trace_data['values']))
        self.assertEqual(10, len(trace_data['iters']))
        self.assertEqual(10, trace_data['iters'][0])
        self.assertEqual(19, trace_data['iters'][-1])

        img_list = trace_data['values'][2]
        self.assertEqual(3, len(img_list))
        img_indices = [img['index'] for img in img_list]
        self.assertListEqual([3, 4, 5], img_indices)


class RunImagesURIBulkLoadApi(RunImagesTestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.image_blobs = {}
        run = Run(run_hash=cls.run_hash, read_only=True)
        empty_context = Context({})
        for step in range(10):
            for idx in range(5):
                img_view = run.series_run_trees[1].subtree(
                    (empty_context.idx, 'random_images', 'val', step, idx)
                )
                cls.image_blobs[img_view['caption']] = img_view['data'].load()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.image_blobs.clear()
        super().tearDownClass()

    def setUp(self) -> None:
        super().setUp()
        self.uri_map = {}
        client = self.client

        response = client.get('/api/runs/search/images/', params={
            'record_range': '0:10',
            'index_range': '0:5',
            'report_progress': False,
        })
        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        run_data = decoded_response[self.run_hash]
        trace_data = run_data['traces'][0]
        for img_list in trace_data['values']:
            for img_data in img_list:
                self.uri_map[img_data['blob_uri']] = img_data['caption']

    def tearDown(self) -> None:
        self.uri_map.clear()
        super().tearDown()

    @parameterized.expand([(1,), (5,), (10,)])
    def test_images_uri_bulk_load_api(self, uri_count):
        # take random N URIs
        uris = random.sample(self.uri_map.keys(), uri_count)

        client = self.client
        response = client.post('/api/runs/images/get-batch', json=uris)
        self.assertEqual(200, response.status_code)
        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(uri_count, len(decoded_response))
        for uri, blob in decoded_response.items():
            expected_blob = self.image_blobs[self.uri_map[uri]]
            self.assertEqual(expected_blob, blob)


class TestRunImagesBatchApi(RunImagesTestBase):
    def test_run_images_bulk_load_api(self):
        client = self.client

        requested_traces = [
            {'name': 'random_images', 'context': {}},
        ]

        response = client.post(f'/api/runs/{self.run_hash}/images/get-batch/', json=requested_traces)
        self.assertEqual(200, response.status_code)

        trace_data = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual('random_images', trace_data['name'])
        self.assertDictEqual({}, trace_data['context'])
        self.assertEqual(51, len(trace_data['values']))
        expected_iters = list(range(0, 100, 2)) + [99]  # include last step
        self.assertListEqual(expected_iters, trace_data['iters'])

        img_list = trace_data['values'][2]
        self.assertEqual(6, len(img_list))

        img = img_list[3]
        self.assertEqual('Image 4 9', img['caption'])
        self.assertEqual(9, img['index'])
        self.assertEqual(16, img['width'])
        self.assertEqual(16, img['height'])
        self.assertEqual('png', img['format'])
        self.assertIn('blob_uri', img)


class TestImageListsAndSingleImagesSearchApi(ApiTestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        run = cls.create_run(system_tracking_interval=None)
        cls.run_hash = run.hash

        for step in range(5):
            images = generate_image_set(img_count=5, caption_prefix=f'Image {step}')
            run.track(images, name='image_lists', context={})
            run.track(images[0], name='single_images', context={})

    def test_search_simgle_image_only_default_index_range(self):
        client = self.client

        query = self.isolated_query_patch('images.name == "single_images"')
        response = client.get('/api/runs/search/images/',
                              params={'q': query, 'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        self.assertEqual([0, 5], run_data['ranges']['record_range_total'])
        self.assertEqual([0, 5], run_data['ranges']['record_range_used'])
        self.assertEqual([0, 1], run_data['ranges']['index_range_total'])
        self.assertEqual([0, 1], run_data['ranges']['index_range_used'])

        self.assertEqual(1, len(run_data['traces']))
        trace_data = run_data['traces'][0]
        self.assertEqual('single_images', trace_data['name'])
        self.assertEqual(5, len(trace_data['values']))
        img_list = trace_data['values'][0]
        self.assertEqual(1, len(img_list))

    def test_mixed_search_default_index_range(self):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/', params={'q': query, 'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        self.assertEqual([0, 5], run_data['ranges']['record_range_total'])
        self.assertEqual([0, 5], run_data['ranges']['record_range_used'])
        self.assertEqual([0, 5], run_data['ranges']['index_range_total'])
        self.assertEqual([0, 5], run_data['ranges']['index_range_used'])

        self.assertEqual(2, len(run_data['traces']))
        trace_data = run_data['traces'][0]
        self.assertEqual('image_lists', trace_data['name'])
        self.assertEqual(5, len(trace_data['values']))
        img_list = trace_data['values'][0]
        self.assertEqual(5, len(img_list))

        trace_data = run_data['traces'][1]
        self.assertEqual('single_images', trace_data['name'])
        self.assertEqual(5, len(trace_data['values']))
        img_list = trace_data['values'][0]
        self.assertEqual(1, len(img_list))

    def test_mixed_search_custom_index_range(self):
        client = self.client

        query = self.isolated_query_patch()
        response = client.get('/api/runs/search/images/',
                              params={'q': query, 'index_range': '3:5', 'report_progress': False})
        self.assertEqual(200, response.status_code)

        decoded_response = decode_tree(decode_encoded_tree_stream(response.iter_content(chunk_size=512 * 1024)))
        self.assertEqual(1, len(decoded_response))
        run_data = decoded_response[self.run_hash]
        self.assertEqual([0, 5], run_data['ranges']['record_range_total'])
        self.assertEqual([0, 5], run_data['ranges']['record_range_used'])
        self.assertEqual([0, 5], run_data['ranges']['index_range_total'])
        self.assertEqual([3, 5], run_data['ranges']['index_range_used'])

        self.assertEqual(2, len(run_data['traces']))
        trace_data = run_data['traces'][0]
        self.assertEqual('image_lists', trace_data['name'])
        self.assertEqual(5, len(trace_data['values']))
        img_list = trace_data['values'][0]
        self.assertEqual(2, len(img_list))

        trace_data = run_data['traces'][1]
        self.assertEqual('single_images', trace_data['name'])
        self.assertEqual(5, len(trace_data['values']))
        img_list = trace_data['values'][0]
        self.assertListEqual([], img_list)


class TestRunInfoApi(ApiTestBase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()

        # run1 -> context {'subset': 'train'} -> Image[]
        #      |                              -> integers
        #      |                              -> floats
        #      -> context {'subset': 'val'}   -> floats
        # ------------------------------------------------
        # run2 -> context {'subset': 'train'} -> Image
        #      |                              -> floats
        #      -> context {'subset': 'val'}   -> floats

        run1 = cls.create_run(system_tracking_interval=None)
        cls.run1_hash = run1.hash
        images = generate_image_set(img_count=2, caption_prefix=f'Image 0')
        run1.track(images, name='image_lists', context={'subset': 'train'})
        run1.track(random.random(), name='floats', context={'subset': 'train'})
        run1.track(random.randint(100, 200), name='integers', context={'subset': 'train'})
        run1.track(random.random(), name='floats', context={'subset': 'val'})

        run2 = cls.create_run(system_tracking_interval=None)
        run2.track(images[0], name='single_images', context={'subset': 'val'})
        run2.track(random.random(), name='floats', context={'subset': 'train'})
        run2.track(random.random(), name='floats', context={'subset': 'val'})
        cls.run2_hash = run2.hash

    def test_run_info_get_images_only_api(self):
        client = self.client
        response = client.get(f'api/runs/{self.run1_hash}/info', params={'sequence': 'images'})
        self.assertEqual(200, response.status_code)
        response_data = response.json()
        self.assertEqual(1, len(response_data['traces']))
        self.assertIn('images', response_data['traces'])
        self.assertDictEqual({'subset': 'train'}, response_data['traces']['images'][0]['context'])
        self.assertEqual('image_lists', response_data['traces']['images'][0]['name'])

        response = client.get(f'api/runs/{self.run2_hash}/info', params={'sequence': 'images'})
        self.assertEqual(200, response.status_code)
        response_data = response.json()
        self.assertEqual(1, len(response_data['traces']))
        self.assertIn('images', response_data['traces'])
        self.assertDictEqual({'subset': 'val'}, response_data['traces']['images'][0]['context'])
        self.assertEqual('single_images', response_data['traces']['images'][0]['name'])

    @parameterized.expand([
        ({'sequence': ('metric', 'images', 'audios',
                       'distributions', 'figures', 'texts', 'logs')}, 7),  # explicit specification
        (None, 8)                                                          # default
    ])
    def test_run_info_get_all_sequences_api(self, qparams, trace_type_count):
        client = self.client
        response = client.get(f'api/runs/{self.run1_hash}/info', params=qparams)
        self.assertEqual(200, response.status_code)
        response_data = response.json()
        self.assertEqual(trace_type_count, len(response_data['traces']))
        self.assertIn('images', response_data['traces'])
        self.assertIn('audios', response_data['traces'])
        self.assertIn('metric', response_data['traces'])
        self.assertIn('distributions', response_data['traces'])
        self.assertIn('figures', response_data['traces'])
        self.assertIn('texts', response_data['traces'])
        self.assertDictEqual({'subset': 'train'}, response_data['traces']['images'][0]['context'])
        self.assertEqual('image_lists', response_data['traces']['images'][0]['name'])
        metrics_data = response_data['traces']['metric']
        self.assertEqual(3, len(metrics_data))
        self.assertEqual('floats', metrics_data[0]['name'])
        self.assertEqual('floats', metrics_data[1]['name'])
        self.assertEqual('integers', metrics_data[2]['name'])
        self.assertDictEqual({'subset': 'val'}, metrics_data[0]['context'])
        self.assertDictEqual({'subset': 'train'}, metrics_data[1]['context'])
        self.assertDictEqual({'subset': 'train'}, metrics_data[2]['context'])

        response = client.get(f'api/runs/{self.run2_hash}/info', params={'sequence': ('images', 'metric')})
        self.assertEqual(200, response.status_code)
        response_data = response.json()
        self.assertEqual(2, len(response_data['traces']))
        self.assertIn('images', response_data['traces'])
        self.assertIn('metric', response_data['traces'])
        self.assertDictEqual({'subset': 'val'}, response_data['traces']['images'][0]['context'])
        self.assertEqual('single_images', response_data['traces']['images'][0]['name'])
        metrics_data = response_data['traces']['metric']
        self.assertEqual(2, len(metrics_data))
        self.assertEqual('floats', metrics_data[0]['name'])
        self.assertEqual('floats', metrics_data[1]['name'])
        self.assertDictEqual({'subset': 'val'}, metrics_data[0]['context'])
        self.assertDictEqual({'subset': 'train'}, metrics_data[1]['context'])

    def test_run_info_get_metrics_only_api(self):
        client = self.client
        response = client.get(f'api/runs/{self.run1_hash}/info', params={'sequence': 'metric'})
        self.assertEqual(200, response.status_code)
        response_data = response.json()
        self.assertEqual(1, len(response_data['traces']))
        self.assertIn('metric', response_data['traces'])
        self.assertEqual(3, len(response_data['traces']['metric']))

        response = client.get(f'api/runs/{self.run2_hash}/info', params={'sequence': 'metric'})
        self.assertEqual(200, response.status_code)
        response_data = response.json()
        self.assertEqual(1, len(response_data['traces']))
        self.assertIn('metric', response_data['traces'])
        self.assertEqual(2, len(response_data['traces']['metric']))

    def test_invalid_sequence_type(self):
        client = self.client
        response = client.get(f'api/runs/{self.run1_hash}/info', params={'sequence': 'non-existing-sequence'})
        self.assertEqual(400, response.status_code)



import logging

from typing import Union, Optional, TYPE_CHECKING

from aim.sdk.base_run import BaseRun
from aim.ext.cleanup import AutoClean

if TYPE_CHECKING:
    from aim.sdk import Repo

logger = logging.getLogger(__name__)


class RunAutoClean(AutoClean['MaintenanceRun']):
    PRIORITY = 90

    def __init__(self, instance: 'MaintenanceRun') -> None:
        """
        Prepare the `Run` for automatic cleanup.

        Args:
            instance: The `Run` instance to be cleaned up.
        """
        super().__init__(instance)

        self._lock = instance._lock

    def _close(self) -> None:
        if self._lock is not None:
            self._lock.release()


class MaintenanceRun(BaseRun):
    def __init__(self, run_hash: str, repo: Optional[Union[str, 'Repo']] = None):
        self._resources: Optional[RunAutoClean] = None
        super().__init__(run_hash, repo=repo, read_only=False)
        self._resources = RunAutoClean(self)



from aim import Repo

from performance_tests.base import StorageTestBase
from performance_tests.utils import get_baseline, write_baseline
from performance_tests.storage.utils import iterative_access_metric_values


class TestIterativeAccessExecutionTime(StorageTestBase):
    def test_iterative_access(self):
        test_name = 'test_iterative_access'
        repo = Repo.default_repo()
        query = 'metric.name == "metric 0"'
        execution_time = iterative_access_metric_values(repo, query)
        baseline = get_baseline(test_name)
        if baseline:
            self.assertInRange(execution_time, baseline)
        else:
            write_baseline(test_name, execution_time)



import io
import logging
import re
import sys
import time
import weakref

from psutil import Process, cpu_percent
from threading import Thread
from typing import Union
from weakref import WeakValueDictionary

from aim.ext.resource.configs import AIM_RESOURCE_METRIC_PREFIX
from aim.ext.resource.log import LogLine
from aim.ext.resource.stat import Stat

logger = logging.getLogger(__name__)


class ResourceTracker(object):
    _buffer_registry = WeakValueDictionary()
    _old_out_write = None
    _old_err_write = None

    @classmethod
    def _install_stream_patches(cls):
        cls._old_out_write = sys.stdout.write
        cls._old_err_write = sys.stderr.write

        def new_out_write(data):
            cls._old_out_write(data)
            if isinstance(data, str):
                data = data.encode()
            for buffer in cls._buffer_registry.values():
                buffer.write(data)

        def new_err_write(data):
            cls._old_err_write(data)
            if isinstance(data, str):
                data = data.encode()
            for buffer in cls._buffer_registry.values():
                buffer.write(data)

        sys.stdout.write = new_out_write
        sys.stderr.write = new_err_write

    @classmethod
    def _uninstall_stream_patches(cls):
        sys.stdout.write = cls._old_out_write
        sys.stderr.write = cls._old_err_write

    STAT_INTERVAL_MIN = 0.1
    STAT_INTERVAL_MAX = 24 * 60 * 60.0
    STAT_INTERVAL_DEFAULT = 60.0

    @classmethod
    def check_interval(cls, interval, warn=True):
        if interval is None:
            warn = False
        if not isinstance(interval, (int, float)) or not cls.STAT_INTERVAL_MIN <= interval <= cls.STAT_INTERVAL_MAX:
            if warn:
                logger.warning('To track system resource usage '
                               'please set `system_tracking_interval` '
                               'greater than 0 and less than 1 day')
            return False
        return True

    reset_cpu_cycle = False

    @staticmethod
    def reset_proc_interval():
        """
        Calls process `cpu_percent` which resets cpu utilization tracking cycle
        Read more: https://psutil.readthedocs.io/en/latest/#psutil.cpu_percent
        """
        cpu_percent(0.0)

    def __init__(self,
                 tracker,
                 interval: Union[int, float] = STAT_INTERVAL_DEFAULT,
                 capture_logs: bool = True,
                 log_offset: int = 0):
        self._tracker = weakref.ref(tracker)
        self._stat_capture_interval = None
        if self.check_interval(interval, warn=False):
            self._stat_capture_interval = interval

        # terminal log capturing
        self._capture_logs = capture_logs
        self._log_capture_interval = 1
        self._old_out = None
        self._old_err = None
        self._io_buffer = io.BytesIO()
        self._line_counter = log_offset

        try:
            self._process = Process()
        except Exception:
            self._process = None

        # Start thread to collect stats and logs at intervals
        self._th_collector = Thread(target=self._stat_collector, daemon=True)
        self._shutdown = False
        self._started = False

        if ResourceTracker.reset_cpu_cycle is False:
            ResourceTracker.reset_cpu_cycle = True
            self.reset_proc_interval()

    def start(self):
        """
        Start statistics collection
        """
        if self._started:
            return

        self._started = True
        if self._capture_logs:
            # install the stream patches if not done yet
            if not self._buffer_registry:
                self._install_stream_patches()
            self._buffer_registry[id(self)] = self._io_buffer
        # Start thread to asynchronously collect statistics
        self._th_collector.start()

    def stop(self):
        if not self._started:
            return

        self._shutdown = True
        self._th_collector.join()
        if self._capture_logs:
            # read and store remaining buffered logs
            self._store_buffered_logs()
            # unregister the buffer
            del self._buffer_registry[id(self)]
            # uninstall stream patching if no buffer is left in the registry
            if not self._buffer_registry:
                self._uninstall_stream_patches()

    def close(self):
        """Interface to make compatible with Resource AutoClean"""
        self.stop()

    def _track(self, stat: Stat):
        # Store system stats
        for resource, usage in stat.system.items():
            self._tracker()(
                usage,
                name='{}{}'.format(AIM_RESOURCE_METRIC_PREFIX, resource),
            )

        # Store GPU stats
        for gpu_idx, gpu in enumerate(stat.gpus):
            for resource, usage in gpu.items():
                self._tracker()(
                    usage,
                    name='{}{}'.format(AIM_RESOURCE_METRIC_PREFIX, resource),
                    context={'gpu': gpu_idx}
                )

    def _stat_collector(self):
        """
            Statistics collecting thread body
        """
        stat_time_counter = 0
        log_capture_time_counter = 0

        # store initial system usage stats
        if self._stat_capture_interval:
            stat = Stat(self._process)
            self._track(stat)

        while True:
            # Get system statistics
            if self._shutdown:
                break

            time.sleep(0.1)
            stat_time_counter += 0.1
            log_capture_time_counter += 0.1

            if self._stat_capture_interval and stat_time_counter > self._stat_capture_interval:
                stat = Stat(self._process)
                self._track(stat)
                stat_time_counter = 0

            if self._capture_logs and log_capture_time_counter > self._log_capture_interval:
                self._store_buffered_logs()
                log_capture_time_counter = 0

    def _store_buffered_logs(self):
        _buffer_size = self._io_buffer.tell()
        if not _buffer_size:
            return

        self._io_buffer.seek(0)
        # read and reset the buffer
        data = self._io_buffer.read(_buffer_size)
        self._io_buffer.seek(0)
        # handle the buffered data and store

        lines = data.split(b'\n')
        ansi_csi_re = re.compile(b"\001?\033\\[((?:\\d|;)*)([a-dA-D])\002?")

        def _handle_csi(line):
            def _remove_csi(line):
                return re.sub(ansi_csi_re, b'', line)

            for match in ansi_csi_re.finditer(line):
                arg, command = match.groups()
                arg = int(arg.decode()) if arg else 1
                if command == b'A':  # cursor up
                    self._line_counter -= arg
                if command == b'B':  # cursor down
                    self._line_counter += arg

            return _remove_csi(line)

        line = None
        for line in lines:
            # handle cursor up and down symbols
            line = _handle_csi(line)
            # handle each line for carriage returns
            line = line.rsplit(b'\r')[-1]
            self._tracker()(LogLine(line.decode()), name='logs', step=self._line_counter)
            self._line_counter += 1

        self._line_counter -= 1

        # if there was no b'\n' at the end of the data keep the last line in buffer for further writing
        if line != b'':
            self._io_buffer.write(line)



from pydantic import BaseModel
from uuid import UUID
from typing import List, Optional


class TagCreateIn(BaseModel):
    name: str
    color: Optional[str] = ''
    description: Optional[str] = ''


class TagUpdateIn(BaseModel):
    name: Optional[str] = ''
    color: Optional[str] = None
    description: Optional[str] = None
    archived: Optional[bool] = None


class TagUpdateOut(BaseModel):
    id: UUID
    status: str = 'OK'


class TagGetOut(BaseModel):
    id: UUID
    name: str
    color: Optional[str] = None
    description: Optional[str] = None
    run_count: int = 0
    archived: bool


TagListOut = List[TagGetOut]


class TagGetRunsOut(BaseModel):
    class Run(BaseModel):
        run_id: str
        name: str
        experiment: Optional[str] = None
        creation_time: float
        end_time: Optional[float]

    id: UUID
    runs: List[Run]



from parameterized import parameterized

from performance_tests.base import SDKTestBase
from performance_tests.utils import (
    get_baseline,
    write_baseline
)
from performance_tests.sdk.queries import queries
from performance_tests.sdk.utils import (
    collect_runs_data,
    collect_metrics_data
)


class TestDataCollectionExecutionTime(SDKTestBase):
    @parameterized.expand(queries.items())
    def test_collect_runs_data(self, query_key, query):
        query_execution_time = collect_runs_data(query)
        test_name = f'test_collect_runs_data_{query_key}'
        baseline = get_baseline(test_name)
        if baseline:
            self.assertInRange(query_execution_time, baseline)
        else:
            write_baseline(test_name, query_execution_time)

    @parameterized.expand(queries.items())
    def test_collect_metrics_data(self, query_key, query):
        query_execution_time = collect_metrics_data(query)
        test_name = f'test_collect_metrics_data_{query_key}'
        baseline = get_baseline(test_name)
        if baseline:
            self.assertInRange(query_execution_time, baseline)
        else:
            write_baseline(test_name, query_execution_time)



from aim.storage.object import CustomObject
from aim.storage.types import BLOB


@CustomObject.alias('aim.text')
class Text(CustomObject):
    """Text object used to store text objects in Aim repository.

        Args:
             text (:obj:): str object used to construct `aim.Text`.
        """

    AIM_NAME = 'aim.text'

    def __init__(self, text):
        super().__init__()

        if not isinstance(text, str):
            raise TypeError('`text` must be a string.')

        self.storage['data'] = BLOB(data=text)

    def json(self):
        return {}

    @property
    def data(self):
        return self.storage['data'].load()

    def __repr__(self):
        return self.data



from typing import Tuple, Union

from aim.sdk.sequence import MediaSequenceBase
from aim.sdk.objects.text import Text


class Texts(MediaSequenceBase):
    """Class representing series of Text objects."""

    @classmethod
    def allowed_dtypes(cls) -> Union[str, Tuple[str, ...]]:
        text_typename = Text.get_typename()
        return text_typename, f'list({text_typename})'

    @classmethod
    def sequence_name(cls) -> str:
        return 'texts'



# distutils: language = c++
# cython: language_level = 3

from aim.storage cimport utils
from aim.storage.utils cimport interfaces

cdef class ContainerItemsIterator(interfaces.Iterator):
    pass

cdef class Container:
    cdef __weakref__

    cpdef void close(self)
    cpdef void preload(self)
    cpdef object get(self, bytes key, object default = *)
    cpdef void set(self, bytes key, object value, store_batch = *)
    cpdef void delete_range(self, bytes begin, bytes end, store_batch = *)

    cpdef ContainerItemsIterator items(self, bytes prefix = *)

    cpdef bytes next_key(self, bytes key = *)
    cpdef object next_value(self, bytes key = *)
    cpdef tuple next_key_value(self, bytes key = *)
    cpdef tuple next_item(self, bytes key = *)
    cpdef bytes prev_key(self, bytes key = *)
    cpdef object prev_value(self, bytes key = *)
    cpdef tuple prev_key_value(self, bytes key = *)
    cpdef tuple prev_item(self, bytes key = *)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
# temporary workaround for M1 build
try:
    import grpc
except ImportError:
    grpc = None

import aim.ext.transport.health_pb2 as health__pb2


class HealthStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Check = channel.unary_unary(
                '/grpc.health.v1.Health/Check',
                request_serializer=health__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=health__pb2.HealthCheckResponse.FromString,
                )
        self.Watch = channel.unary_stream(
                '/grpc.health.v1.Health/Watch',
                request_serializer=health__pb2.HealthCheckRequest.SerializeToString,
                response_deserializer=health__pb2.HealthCheckResponse.FromString,
                )


class HealthServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Check(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Watch(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_HealthServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Check': grpc.unary_unary_rpc_method_handler(
                    servicer.Check,
                    request_deserializer=health__pb2.HealthCheckRequest.FromString,
                    response_serializer=health__pb2.HealthCheckResponse.SerializeToString,
            ),
            'Watch': grpc.unary_stream_rpc_method_handler(
                    servicer.Watch,
                    request_deserializer=health__pb2.HealthCheckRequest.FromString,
                    response_serializer=health__pb2.HealthCheckResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'grpc.health.v1.Health', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class Health(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Check(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/grpc.health.v1.Health/Check',
            health__pb2.HealthCheckRequest.SerializeToString,
            health__pb2.HealthCheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Watch(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/grpc.health.v1.Health/Watch',
            health__pb2.HealthCheckRequest.SerializeToString,
            health__pb2.HealthCheckResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



# distutils: language = c++
# cython: language_level = 3

cimport aimrocks.lib_rocksdb as interfaces

cdef class ArrayFlagType:
    pass

cdef class ObjectFlagType:
    pass

cdef class CustomObjectFlagType:
    cdef:
        public str aim_name

cdef class BLOB:
    cdef:
        object data
        object loader_fn
    cpdef object load(self)
    # TODO closures inside cython functions are not supported yet
    # cpdef object transform(self, object transform_fn)



import logging

from abc import abstractmethod
from functools import lru_cache
from datetime import datetime, timedelta

from RestrictedPython import (
    safe_builtins,
    utility_builtins,
    limited_builtins,
    compile_restricted
)
from RestrictedPython.Eval import default_guarded_getitem
from RestrictedPython.Guards import (
    full_write_guard,
    guarded_iter_unpack_sequence,
    guarded_unpack_sequence
)


def safe_import(*args, **kwargs):
    if args and args[0] != 'time':
        raise ImportError(f'{args[0]} package cannot be imported.')
    return __import__(*args, **kwargs)


extra_builtins = {
    'datetime': datetime,
    'timedelta': timedelta,
    'sorted': sorted,
    'min': min,
    'max': max,
    'sum': sum,
    'any': any,
    'all': all,
    '__import__': safe_import,
}

builtins = safe_builtins.copy()
builtins.update(utility_builtins)
builtins.update(limited_builtins)
builtins.update(extra_builtins)


def safer_getattr(object, name, default=None, getattr=getattr):
    '''Getattr implementation which prevents using format on string objects.

    format() is considered harmful:
    http://lucumr.pocoo.org/2016/12/29/careful-with-str-format/

    '''
    if name == 'format' and isinstance(object, str):
        raise NotImplementedError(
            'Using format() on a %s is not safe.' % object.__class__.__name__)
    if name[0] == '_':
        raise AttributeError(
            '"{name}" is an invalid attribute name because it '
            'starts with "_"'.format(name=name)
        )
    val = getattr(object, name, default)
    return val


restricted_globals = {
    "__builtins__": builtins,
    "_getattr_": safer_getattr,
    "_write_": full_write_guard,
    "_getiter_": iter,
    "_getitem_": default_guarded_getitem,
    "_iter_unpack_sequence_": guarded_iter_unpack_sequence,
    "_unpack_sequence_": guarded_unpack_sequence
}

logger = logging.getLogger(__name__)


# CODE_FORMAT = """{expr})"""


class Query:

    __slots__ = ('__weakref__', 'expr')

    def __init__(
        self,
        expr: str
    ):
        self.expr = expr

    @abstractmethod
    def check(
        self,
        **params
    ) -> bool:
        ...

    def __call__(
        self,
        **params
    ):
        return self.check(**params)


@lru_cache(maxsize=100)
def compile_checker(expr):
    source_code = expr
    byte_code = compile_restricted(source_code,
                                   filename='<inline code>',
                                   mode='eval')
    return byte_code


def syntax_error_check(expr):
    if not expr:
        return
    expr = strip_query(expr)
    try:
        compile_restricted(expr,
                           filename='<inline code>',
                           mode='eval')
    except SyntaxError:
        compile(expr, filename='<inline code>', mode='eval')


@lru_cache(maxsize=100)
def strip_query(query: str) -> str:
    import re
    stripped_query = query.strip()
    # cut the hardcoded part (SELECT something IF)
    if query.lower().startswith('select'):
        try:
            stripped_query = re.split('if',
                                      query,
                                      maxsplit=1,
                                      flags=re.IGNORECASE)[1]
        except IndexError:
            stripped_query = ''

    if stripped_query:
        stripped_query = f'({stripped_query.strip()})'

    return stripped_query


@lru_cache(maxsize=100)
def query_add_default_expr(query: str) -> str:
    default_expression = 'run.archived == False'
    # add the default expression to the query if needed
    if not query:
        return default_expression
    else:
        if 'run.archived' not in query:
            return f'{default_expression} and {query}'
        else:
            return query


class RestrictedPythonQuery(Query):

    __slots__ = ('_checker', 'run_metadata_cache')

    allowed_params = {'run', 'metric', 'images', 'audios', 'distributions', 'figures', 'texts'}

    def __init__(
        self,
        query: str
    ):
        stripped_query = strip_query(query)
        expr = query_add_default_expr(stripped_query)
        super().__init__(expr=expr)
        self._checker = compile_checker(expr)
        self.run_metadata_cache = None

    def __bool__(
        self
    ) -> bool:
        return bool(self.expr)

    def check(
        self,
        **params
    ) -> bool:
        # prevent possible messing with globals
        assert set(params.keys()).issubset(self.allowed_params)

        # TODO enforce immutable
        try:
            namespace = dict(**params, **restricted_globals)
            return eval(self._checker, restricted_globals, namespace)
        except BaseException as e:
            logger.warning('query failed, %s', e)
            return False



from aim.ext.transport.message_utils import ResourceObject, pack_args
from aim.ext.transport.remote_resource import RemoteResourceAutoClean

from aim.storage.treeview import TreeView
from aim.storage.treeutils import encode_tree
from aim.storage.treearrayview import TreeArrayView
from aim.storage.types import AimObject, AimObjectKey, AimObjectPath

from typing import TYPE_CHECKING, Any, Iterator, Tuple, Union, List


if TYPE_CHECKING:
    from aim.ext.transport.client import Client


class ProxyTreeAutoClean(RemoteResourceAutoClean):
    PRIORITY = 60


class ProxyTree(TreeView):
    def __init__(self, client: 'Client',
                 name: str,
                 sub: str,
                 read_only: bool,
                 from_union: bool = False,
                 index=False,
                 timeout=None):
        self._resources: ProxyTreeAutoClean = None

        self._rpc_client = client
        self._hash = sub

        kwargs = {
            'name': name,
            'sub': sub,
            'read_only': read_only,
            'from_union': from_union,
            'index': index,
            'timeout': timeout,
        }
        self.init_args = pack_args(encode_tree(kwargs))
        self.resource_type = 'TreeView'
        handler = self._rpc_client.get_resource_handler(self, self.resource_type, args=self.init_args)

        self._resources = ProxyTreeAutoClean(self)
        self._resources.rpc_client = client
        self._resources.handler = handler
        self._handler = handler

    def preload(self):
        self._rpc_client.run_instruction(self._hash, self._handler, 'preload')

    def view(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        resolve: bool = False
    ):
        if resolve:
            return None
        # TODO [AT, MV] handle resolve=True
        # make an rpc call to get underlying object type
        # and construct CustomObject if needed
        return SubtreeView(self, path)

    def make_array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ):
        self._rpc_client.run_instruction(self._hash, self._handler, 'make_array', (path,), is_write_only=True)

    def collect(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        strict: bool = True,
        resolve_objects: bool = False
    ) -> AimObject:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'collect', (path, strict, resolve_objects))

    def __delitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath]
    ):
        self._rpc_client.run_instruction(self._hash, self._handler, '__delitem__', (path,), is_write_only=True)

    def __setitem__(
            self,
            path: Union[AimObjectKey, AimObjectPath],
            value: AimObject
    ):
        self._rpc_client.run_instruction(self._hash, self._handler, '__setitem__', (path, value), is_write_only=True)

    def keys_eager(
            self,
            path: Union[AimObjectKey, AimObjectPath] = (),
    ) -> List[Union[AimObjectPath, AimObjectKey]]:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'keys_eager', (path,))

    def keys(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = None
    ) -> List[Union[AimObjectPath, AimObjectKey]]:
        return self.keys_eager(path)

    def items_eager(
            self,
            path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> List[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'items_eager', (path,))

    def items(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Iterator[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        return self.items_eager(path)

    def iterlevel(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = 1
    ) -> Iterator[Tuple[
        AimObjectPath,
        AimObject
    ]]:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'iterlevel', (path, level))

    def array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        dtype: Any = None
    ) -> TreeArrayView:
        return TreeArrayView(self.subtree(path), dtype=dtype)

    def first_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'first_key', (path,))

    def last_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        return self._rpc_client.run_instruction(self._hash, self._handler, 'last_key', (path,))

    def finalize(
        self,
        index: 'ProxyTree'
    ):
        self._rpc_client.run_instruction(self._hash, self._handler, 'finalize', (ResourceObject(index._handler),))


class SubtreeView(TreeView):
    def __init__(self, tree: TreeView, path: Union[AimObjectKey, AimObjectPath]):
        self.tree = tree
        if path == Ellipsis:
            path = ()
        if not isinstance(path, (tuple, list)):
            path = (path,)
        self.prefix = path

    def absolute_path(self, path):
        if path == Ellipsis:
            path = ()
        if not isinstance(path, (tuple, list)):
            path = (path,)
        return self.prefix + path

    def preload(self):
        self.tree.preload()

    def view(
        self,
        path: Union[AimObjectKey, AimObjectPath],
        resolve: bool = False
    ):
        return self.tree.view(self.absolute_path(path), resolve)

    def make_array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ):
        self.tree.make_array(self.absolute_path(path))

    def collect(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        strict: bool = True,
        resolve_objects: bool = False
    ) -> AimObject:
        return self.tree.collect(self.absolute_path(path), strict, resolve_objects)

    def __delitem__(
        self,
        path: Union[AimObjectKey, AimObjectPath]
    ):
        del self.tree[self.absolute_path(path)]

    def __setitem__(
            self,
            path: Union[AimObjectKey, AimObjectPath],
            value: AimObject
    ):
        self.tree[self.absolute_path(path)] = value

    def keys(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = None
    ) -> Iterator[Union[AimObjectPath, AimObjectKey]]:
        return self.tree.keys(self.absolute_path(path))

    def items(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> Iterator[Tuple[
        AimObjectKey,
        AimObject
    ]]:
        return self.tree.items(self.absolute_path(path))

    def iterlevel(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        level: int = 1
    ) -> Iterator[Tuple[
        AimObjectPath,
        AimObject
    ]]:
        return self.tree.iterlevel(self.absolute_path(path), level)

    def array(
        self,
        path: Union[AimObjectKey, AimObjectPath] = (),
        dtype: Any = None
    ) -> TreeArrayView:
        return TreeArrayView(self.subtree(path), dtype=dtype)

    def first_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        return self.tree.first_key(self.absolute_path(path))

    def last_key(
        self,
        path: Union[AimObjectKey, AimObjectPath] = ()
    ) -> AimObjectKey:
        return self.tree.last_key(self.absolute_path(path))

    def finalize(
        self,
        index: 'SubtreeView'
    ):
        self.tree.finalize(index=index.tree)



"""empty message

Revision ID: 46b89d830ad8
Revises: b07e7b07c8ce
Create Date: 2022-11-09 05:05:48.794990

"""
from alembic import op
import sqlalchemy as sa


# revision identifiers, used by Alembic.
revision = '46b89d830ad8'
down_revision = 'b07e7b07c8ce'
branch_labels = None
depends_on = None


def upgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.add_column('experiment', sa.Column('description', sa.Text(), nullable=True))
    op.add_column('note', sa.Column('experiment_id', sa.Integer(), nullable=True))

    with op.batch_alter_table('note') as batch_op:
        batch_op.create_foreign_key('fk_experiment_note', 'experiment', ['experiment_id'], ['id'])

    op.create_table('run_info',
    sa.Column('id', sa.Integer(), autoincrement=True, nullable=False),
    sa.Column('run_id', sa.Integer(), nullable=True),
    sa.Column('created_at', sa.DateTime(), nullable=True),
    sa.Column('updated_at', sa.DateTime(), nullable=True),
    sa.Column('last_notification_index', sa.Integer(), nullable=True),
    sa.ForeignKeyConstraint(['run_id'], ['run.id'], ),
    sa.PrimaryKeyConstraint('id')
    )
    # ### end Alembic commands ###


def downgrade():
    # ### commands auto generated by Alembic - please adjust! ###
    op.drop_table('run_info')

    op.drop_constraint(None, 'note', type_='foreignkey')
    op.drop_column('note', 'experiment_id')
    op.drop_column('experiment', 'description')
    # ### end Alembic commands ###



"""
This system is designed to let users to check if the run had any progress.

While it may seem an easy task, it's very difficult to have a reliable way to
check if everything is fine with the process; The main (rank0) process may be
killed having no chance of notifying about failure. Even if the main process is
alive, the other moving parts may be stuck (network sync with other nodes,
filesystem, dataloader) those are not necessarily part of the main execution
thread. It leaves no other choice other than defining:

> The run is considered to be failed if it
> had not reported any progress in the promised time.

Thus, we can mark progress (we call these `check-in`s) and the expected time of
the next check-in. We can report a check-in after each forwarding each batch and
promising the next check-in to be reported in the next minute. If no check-ins
are received in the next minute, the run is most certainly failed or stuck (or
very unexpected things happened thus compromising the performance of the system).
This can have some exceptions: for instance, checkpointing the model and saving
the state into the filesystem may take a while. For such cases, aim integration
should check-in right before triggering checkpointing by promising much larger
time intervals.

Monitoring of run progress can be easily done by a service that periodically
polls the check-in instances. If the last check-in was there for longer than it
promised, the run is considered to  be failed thus triggering a (configurable)
failure notification.

Check-ins with zero `expect_next_in` values denote an absence of the expiration
date. In order to mark the run as successful, the last check-in should be
reported with a zero `expect_next_in` to indicate no further check-ins are
expected so the monitoring server will not treat stopped process as failed.

# IMPLEMENTATION DETAILS

## FORMAT
The check-ins are implemented to be stored in the filesystem, in `.aim` repo
under the `check_ins` directory. Each file has no contents and encodes all the
information in the name.

The name format is as follows:
    `{run_hash}-{idx:08d}-{flag_name}-{absolute_time:011.2f}-{expect_next_in:05d}`
where `idx` is auto-incremented index per run, `flag_name` is the name of the
flag (usually `check-in`), `absolute_time` is the utc time of the check-in (is
intended only for debugging purposes), `expect_next_in` is the time in seconds
until the next check-in is expected.

## TIME
Note that we don't use last modified time to detect if the file is modified
because it's not that reliable especially for virtual and remote filesystems.
The utc time in the filename may vary across machines, and used only for the
debugging purposes.

## MONITORING SERVICE
The monitoring service supposed to periodically poll the directory and check for
the latest (lexicographically highest) check-ins per run.
If the last check-in was there (starting from the first time the monitoring
server had seen the file) for longer than it promised, the run is considered to
be failed thus triggering a (configurable) failure notification. A grace period is
introduced to avoid false positives.

## NON-BLOCKING INTERFACE

### `RunStatusReporter.report_progress()`:
By default, the check-in is non-blocking call in order to avoid any latency
introduced. This way, the caller can feel free to check-in as soon as possible,
even after each batch is forwarded or report the progress in vert small steps.

In order to avoid overloading the filesystem, not all the check-ins are stored
in the filesystem. Instead, if there is still time before the expiration, we can
wait thus avoiding unnecessary disk I/O. The filesystem writes are done only
when the time is (nearly) up. A separate thread is used to handle this process
which also cleans up of the obsolete check-ins having lower lexicographic order
than the current one.

### `RunStatusReporter.report_successful_finish()`:
Marking the run as successful is blocking call by default to ensure the check-in
is written to the filesystem before the run process exits. Note, that the
initialization of `RunStatusReporter` also does include blocking call to poll the
latest state in case of existing runs.

## INDIRECT INTERFACE
The preferred way of reporting check-ins is to call `.report_progress()` method on the
Run instance. However, in certain cases, the caller may want to check-in from a
code location that has no access to the Run instance (e.g. in dataloader, or in
the checkpoint callback). In such cases, the `.report_progress()` method can be called
indirectly by calling `RunStatusReporter.report_progress()` which will infer the run instance.

This is only possible if there is a single run instance in the process.


# TODO [mahnerak] Add Revision on:
  - [x] Multi-flag support, independent and global flushes
  - [x] Choosing stricter `expect_next_in` values
  - [ ] Legacy support for `progress` check-ins.
"""

import math
import time
import queue
import threading
from collections import defaultdict
from dataclasses import dataclass, field
from pathlib import Path
from typing import ClassVar, Dict, Optional, Tuple, Set, Union, TYPE_CHECKING

from cachetools import LRUCache

if TYPE_CHECKING:
    from aim.sdk import Run

import logging

logger = logging.getLogger(__name__)


cache = LRUCache(maxsize=3)

# expressed in seconds:
IMMEDIATELY = 0
MIN_SUSPEND_TIME = 0.1
PLAN_ADVANCE_TIME = 10
MAX_SUSPEND_TIME = 30
GRACE_PERIOD = 100
INFINITE_TIME = 1_000_000_000


@dataclass(frozen=True)
class AsteriskType:
    """
    The purpose of this helper symbol is to enable the injecting an asterisk
    into any formatted string no matter what format spec uses.

    For example:
    >>> '{a:.2f}-{b:.2f}'.format(a=3.5, b=Asterisk)
    >>> '3.50-*'
    """

    def __format__(self, __format_spec: str) -> str:
        return "*"


# Singletone instance of AsteriskType
Asterisk = AsteriskType()


@dataclass(order=True, frozen=True)
class CheckIn:
    """
    CheckIn is a dataclass that represents a single check-in.

    All the time-related fields are in `time.monotonic`, relative time expressed in
    seconds, until stated otherwise. In case of `expect_next_in=0` we treat it as
    a special case, meaning that the check-in is not expected to expire.
    Indicies are used to identify check-ins and are unique within a run, auto-incremented.
    """

    idx: int = 0
    expect_next_in: int = field(default=0, compare=False)
    first_seen: float = field(default_factory=time.monotonic, compare=False, repr=False)
    flag_name: str = field(default="check_in", compare=False, repr=True)

    # We keep per-run cache to memoize the first time we've seen check-ins
    per_run_cache: ClassVar[Dict[str, LRUCache]] = defaultdict(
        lambda: LRUCache(maxsize=3)
    )

    def __bool__(self) -> bool:
        return bool(self.idx)

    @classmethod
    def first_seen_cached(
        cls,
        *,
        run_hash: str,
        idx: int,
        expect_next_in: int,
        flag_name: str,
    ) -> float:
        """
        Return the first seen time of the check-in.
        """
        per_run_cache = cls.per_run_cache[run_hash]
        key = (idx, expect_next_in, flag_name)
        try:
            return per_run_cache[key]
        except KeyError:
            now = time.monotonic()
            logger.debug(f"* first time ({now}) seen for {key} for {run_hash}")
            per_run_cache[key] = now
            return now

    @property
    def expiry_date(self) -> float:
        """
        Return the time when the check-in is expected to expire.
        """
        # if the check-in is not expected to expire, we treat it as infinite
        expect_next_in = self.expect_next_in or INFINITE_TIME
        return self.first_seen + expect_next_in

    @classmethod
    def parse(cls, path: Union[Path, str]) -> Tuple[str, "CheckIn"]:
        """
        Parse a check-in file path and return a tuple of:
          * run hash the check-in belongs to,
          * the check-in itself.
        """
        if isinstance(path, str):
            path = Path(path)

        run_hash, str_idx, flag_name, utc_time, str_expect_next_in = path.name.rsplit(
            "-", maxsplit=4
        )

        idx = int(str_idx)
        expect_next_in = int(str_expect_next_in)

        first_seen = cls.first_seen_cached(
            run_hash=run_hash,
            idx=idx,
            expect_next_in=expect_next_in,
            flag_name=flag_name,
        )

        return run_hash, cls(
            idx=idx,
            expect_next_in=expect_next_in,
            first_seen=first_seen,
            flag_name=flag_name,
        )

    @classmethod
    def poll(cls, *, directory: Path, run_hash: str) -> "CheckIn":
        """
        Poll the directory for the current check-in.

        The `current` check-in is defined as the highest one in the lexicographic order.
        """
        pattern = cls.generate_filename(run_hash=run_hash)
        paths = list(directory.glob(pattern))

        if not paths:
            logger.debug(f"no check-in found for {run_hash}; returning zero-check-in")
            return CheckIn()

        check_in_path = max(paths)
        logger.debug(f"found check-in: {check_in_path}")
        parsed_run_hash, check_in = cls.parse(check_in_path)
        assert parsed_run_hash == run_hash
        logger.debug(f"parsed check-in: {check_in}")

        return check_in

    def time_calibrated(self):
        """
        The check-in `expect_next_in` is counted from the time the check-in was
        first created / seen.

        As the `CheckIn` objects are designed to be immutable, we need to
        recalibrate the time when the check-in is expected to expire, returning
        a new `CheckIn` object.
        """
        now = time.monotonic()
        if self.expect_next_in == 0:
            expect_next_in = 0
        else:
            expect_next_in = max(1, math.ceil(self.expiry_date - now))
        new = CheckIn(
            idx=self.idx,
            expect_next_in=expect_next_in,
            first_seen=now,
            flag_name=self.flag_name,
        )
        logger.debug(f"calibrated check-in: {self} -> {new}")
        return new

    @classmethod
    def generate_filename(
        cls,
        *,
        run_hash: Union[str, AsteriskType],
        idx: Union[int, AsteriskType] = Asterisk,
        flag_name: Union[str, AsteriskType] = Asterisk,
        absolute_time: Union[float, AsteriskType] = Asterisk,
        expect_next_in: Union[int, AsteriskType] = Asterisk,
    ) -> str:
        """
        Generate a filename for a check-in.

        Supports Asterisk objects so that we can generate a patterns.

        For example:
        >>> CheckIn.generate_filename(run_hash="run_hash", idx=Asterisk, absolute_time=Asterisk, expect_next_in=10)
        >>> "run_hash-*-check_in-*-10"
        """
        return f"{run_hash}-{idx:016d}-{flag_name}-{absolute_time:011.2f}-{expect_next_in:05d}"

    def cleanup(
        self,
        *,
        directory: Path,
        run_hash: str,
    ) -> "CheckIn":
        """
        Cleanups all the expired check-ins for the given run hash.

        This will remove all check-ins that are older than the current one
        matching the same flag_name.
        Returns the new current check-in.
        """
        pattern = self.generate_filename(run_hash=run_hash, flag_name=self.flag_name)
        *paths_to_remove, current_check_in_path = sorted(directory.glob(pattern))
        logger.debug(f"found {len(paths_to_remove)} check-ins:")
        logger.debug(f"the acting one: {current_check_in_path}")
        for path in paths_to_remove:
            logger.debug(f"check-in {path} is being removed")
            try:
                # Ignore errors, as the file may have been removed already.
                path.unlink()
            except OSError:
                pass
            logger.debug(f"check-in {path} removed")

        parsed_run_hash, check_in = self.parse(current_check_in_path)
        assert parsed_run_hash == run_hash
        logger.debug(f"returning acting check-in after cleanup: {check_in}")

        return check_in

    def touch(
        self,
        *,
        directory: Union[Path, str],
        run_hash: str,
        cleanup: bool = True,
        calibrate: bool = True,
    ) -> "CheckIn":
        """
        Physically write the check-in to the filesystem so that it is visible to the
        other nodes that share the filesystem.

        As `CheckIn.touch()` could be called much later that it is created, we
        need to calibrate `expected_next_in` w.r.t. current time. If `calibrate` is
        `False` (`True` by default), then `expected_next_in` is not recalibrated.

        If `cleanup` is True (default), this will remove all the expired check-ins.
        """
        if not isinstance(directory, Path):
            directory = Path(directory)

        if calibrate and self.expect_next_in:
            return self.time_calibrated().touch(
                directory=directory,
                run_hash=run_hash,
                cleanup=cleanup,
                calibrate=False,
            )

        directory.mkdir(parents=True, exist_ok=True)

        utc_time = time.time()
        filename = self.generate_filename(
            run_hash=run_hash,
            idx=self.idx,
            expect_next_in=self.expect_next_in,
            absolute_time=utc_time,
            flag_name=self.flag_name,
        )
        new_path = directory / filename
        logger.debug(f"touching check-in: {new_path}")

        new_path.touch(exist_ok=True)

        if cleanup:
            self.cleanup(directory=directory, run_hash=run_hash)

        return self

    def time_left(self) -> float:
        """
        Returns the time left before the check-in expires.

        If the check-in has no expiry date, then this returns +infinity.
        """
        if self.expect_next_in == 0:
            return math.inf
        now = time.monotonic()
        return self.expiry_date - now


@dataclass(order=True)
class TimedTask:
    """
    A task for the `TimedQueue` that is scheduled to be executed at a given time.
    Only the `time` field is used for comparison. The `overwritten` field is
    to mock deletion of the task from the priority queue.
    """
    when: float = field(compare=True)
    flag_name: str = field(compare=False)
    overwritten: bool = field(default=False, compare=False)


class RunStatusReporter:
    """
    A handler for check-ins for a given run.

    Uses filesystem to store the check-ins. Handles resumed `Run`s.
    Background thread is used to adjust the rate of check-ins written
    to the filesystem in order to avoid overloading.

    Calling `report_successful_finish()` is required to mark the run as
    finished, otherwise the run will be marked as failed.
    """

    instances: Set["RunStatusReporter"] = set()

    def __init__(
        self,
        run_hash: str,
        repo_path: str,
    ) -> None:
        logger.info(f"creating RunStatusReporter for {run_hash}")
        self.run_hash = run_hash
        self.repo_dir = Path(repo_path)
        self.dir = self.repo_dir / "check_ins"
        logger.debug(f"polling for check-ins in {self.dir}")

        # Detect if the run was resumed. If so, we need to find the last
        # check-in index and continue counting from there.
        leftover = CheckIn.poll(
            directory=self.dir,
            run_hash=self.run_hash,
        )
        if leftover:
            logger.debug(f"leftover check-in: {leftover}")
        else:
            logger.debug("no leftover check-in found. starting from zero")

        self.refresh_condition = threading.Condition()
        self.flush_condition = threading.Condition()
        self.stop_signal = threading.Event()

        # A single lock is used to syncronize all the check-ins API calls.
        # (calls are non-blocking so no need to worry about the performance)
        self.reporter_lock = threading.RLock()
        self.counter = leftover.idx

        # A shared queue for all flags and check-ins is used to avoid having a
        # separate thread for each flag. The queue is sorted by the time when
        # the check-in is expected to be expired.
        #
        # The `last_check_ins` is used to store the last check-in for each flag
        # while `physical_check_ins` is used to store the check-ins that are
        # already written to the filesystem. These are designed to be eventually
        # consistent.
        #
        # The Priority Queue implementation does not allow to update the priority
        # of the existing items, so we need to remove the item and re-insert it
        # with the new priority. This is done by setting `overwritten` flag to
        # `True` and inserting the new item with the same `flag_name`.
        # The `timed_tasks` is used to find the older items to set `overwritten`
        self.queue = queue.PriorityQueue()
        self.last_check_ins: Dict[str, CheckIn] = dict()
        self.physical_check_ins: Dict[str, CheckIn] = dict()
        self.timed_tasks: Dict[str, TimedTask] = dict()

        # We always set the first check-in `starting`.
        logger.info(f"starting from: {self.last_check_ins}")
        self._increment(flag_name="starting")
        # Initialize the background thread and flush all the initial flags
        logger.info(f"starting writer thread for {self}")
        self.thread = threading.Thread(target=self._writer, daemon=True)
        self.thread.start()
        self.flush(block=True)

        # The code below is a workaround to make it accessible from the class
        # interface. This can be especially useful for reporting progress from
        # anywhere in the code.
        self.instances.add(self)
        # We need to patch certain methods to make them
        # available both as instance and class methods.
        self.check_in = self._check_in
        self.report_successful_finish = self._report_successful_finish

    def _touch_flag(
        self,
        flag_name: str,
    ):
        """
        Touches the check-in file on the filesystem. This is blocking call.
        This will also queue the check-in for the next flush if necessary.
        """
        check_in = self.last_check_ins[flag_name]

        if flag_name in self.physical_check_ins and self.physical_check_ins[flag_name].idx == check_in.idx:
            logger.debug(f"Skipping touching `{flag_name}` as it is already up to date")
            self.timed_tasks.pop(flag_name, None)
            logger.debug(f'{flag_name} is not scheduled anymore. It will be invoked as soon as next one appears')
            return

        # Now let's touch the file on the filesystem
        new_check_in = check_in.touch(
            directory=self.dir,
            run_hash=self.run_hash,
        )

        self.physical_check_ins[flag_name] = new_check_in

        # We will take the target (recommended) flush time much shorter.
        expiry_date = min(
            new_check_in.expiry_date,
            time.monotonic() + MAX_SUSPEND_TIME
        )
        # Now we reschedule the check-in for the next flush.
        new_timed_task = TimedTask(
            when=expiry_date,
            flag_name=flag_name,
        )
        self._schedule(new_timed_task)
        return new_check_in

    def _schedule(
        self,
        timed_task: TimedTask,
    ):
        """
        Schedules the check-in for the next flush.
        """
        self.timed_tasks[timed_task.flag_name] = timed_task
        self.queue.put(timed_task)
        with self.refresh_condition:
            self.refresh_condition.notify()

    def _increment(
        self,
        *,
        expect_next_in: int = 0,
        flag_name: str,
    ) -> CheckIn:
        """
        Increments the check-in counter and returns the new check-in for the
        provided flag name. The indices are shared across all the flags.
        This is an (almost) non-blocking call.
        """
        idx = self.counter + 1
        self.counter = idx

        logger.debug(f"incrementing {flag_name} idx -> {idx}")

        check_in = CheckIn(
            idx=idx,
            expect_next_in=expect_next_in,
            flag_name=flag_name
        )

        self.last_check_ins[flag_name] = check_in
        # * If there was no such check-in with the same flag name, we will
        #   schedule it to flush immediately.
        # * We want to make sure that more strict `expect_next_in`s are respected
        #   so we need to reschedule the check-in for the next flush.
        #   If we find much sooner expiry date, we will take it immediately.
        # * Notice that `Immediately` does not mean that the call is blocking.
        was_scheduled = self.timed_tasks.get(flag_name)
        if was_scheduled is None:
            # Schedule to flush ASAP
            timed_task = TimedTask(when=0, flag_name=flag_name)
            self._schedule(timed_task)
            logger.debug(f"scheduled {timed_task} ASAP because no physical check-in was found")
        else:
            if was_scheduled.when > check_in.expiry_date:
                # Schedule to flush ASAP
                timed_task = TimedTask(when=0, flag_name=flag_name)
                self._schedule(timed_task)
                # We need to remove the old task from the queue but the queue
                # implementation does not allow to update the priority of the
                # existing items, so we need to remove the item and re-insert it
                # with the new priority. This is done by setting `overwritten`
                # flag to `True` and inserting the new item with the same
                # `flag_name`. The writer thread will just ignore such items.
                was_scheduled.overwritten = True
                logger.debug(f"scheduled {timed_task} because it newer is stricter than {was_scheduled}")

        return check_in

    @classmethod
    def default(cls) -> "RunStatusReporter":
        """
        Returns the default instance of the RunStatusReporter to make it possible
        to access the instance methods from the class interface. This is useful
        for reporting progress from anywhere in the code.
        """
        try:
            default_instance, = cls.instances
        except ValueError:
            if cls.instances:
                raise ValueError(
                    f"{cls.__name__} has {len(cls.instances)} instances, indirect check-in is not supported"
                )
            else:
                logger.warning(f"{cls.__name__} has no instances")
                return None
        return default_instance

    def close(self):
        """
        Stops the background thread, flushes all the pending check-ins and
        cleans up the resources.
        """
        self.instances.remove(self)
        self.stop()

    def stop(self):
        """
        Disables any new check-ins to be scheduled, flushes all the pending
        check-ins and stops the background thread.
        """
        with self.reporter_lock:
            self.stop_signal.set()
            self.flush(block=True)
            self.thread.join()

    def _writer(self):
        """
        The writer thread that periodically flushes the last check-in.
        This background thread also takes care of cleaning up expired check-ins
        and throttling the check-in rate in order to avoid overloading the
        filesystem. However, it can be notified to flush the last check-in
        on-demand. This is useful when reporting certain events that are preferable
        to be reported as soon as possible, e.g. success status of the Run.
        """

        remaining = 0
        while True:
            # First we need to decide how much time we need to wait to reach the
            # next appropriate flush time. It is either the remaining time from
            # the previous flush, or the moment new check-in is registered.
            with self.refresh_condition:
                logger.debug(f"no interesting things to do, sleeping for {remaining}")
                logger.debug("until woken up")
                logger.debug(f'unfinished tasks: {self.queue.unfinished_tasks}')
                self.refresh_condition.wait(timeout=remaining)

            timed_task: Optional[TimedTask]
            try:
                timed_task = self.queue.get(timeout=0)
            except queue.Empty:
                timed_task = None
            else:
                assert isinstance(timed_task, TimedTask)

                if timed_task.overwritten:
                    logger.debug('detected overwritten task... done')
                    self.queue.task_done()
                    continue

                remaining = timed_task.when - time.monotonic() - PLAN_ADVANCE_TIME
                # remaining = max(remaining, MIN_SUSPEND_TIME)
                remaining = min(remaining, MAX_SUSPEND_TIME)
                logger.debug(f'time remaining: {remaining}')

                if remaining > 0:
                    # TODO Should we push a little late?
                    self._schedule(timed_task)
                    logger.debug(f'too soon, {remaining} remaining')
                    logger.debug(f'putting back for the future: {timed_task}')
                    logger.debug(f'now: {time.monotonic()}... scheduled for: {timed_task.when}')
                    self.queue.task_done()
                    # Mark the task to signal about the clean state.
                    timed_task = None

            # If during the last iteration we could not find a new check-in
            # to flush, we define it as clean state. In the case of if the stop
            # signal is set, we can safely exit the thread as there will be no
            # more check-ins to process in the future.
            if timed_task is None:
                remaining = MAX_SUSPEND_TIME
                with self.flush_condition:
                    self.flush_condition.notify()
                if self.stop_signal.is_set():
                    return
            else:
                logger.debug(f'only {remaining} remaining... flushing one task')
                self._touch_flag(timed_task.flag_name)
                self.queue.task_done()
                # Let's immediately proceed to the next iteration to check if
                # there are any new check-ins to flush.
                remaining = IMMEDIATELY

    def flush(
        self,
        flag_name: Optional[str] = None,
        block: bool = True,
    ) -> None:
        """
        Flush the last check-in.
        If `flag_name` is specified, only the check-ins for the given flag
        will be flushed.
        Otherwise, all the check-ins will be flushed. In this case, the order
        of (active) check-ins (per flag name) will be preserved.
        """
        logger.debug(f"notifying {self}")

        with self.reporter_lock:
            flag_names = [flag_name] if flag_name is not None else self.timed_tasks
            with self.flush_condition:
                for flag_name in flag_names:
                    logger.debug(f"flushing {flag_name}")
                    # We add a new task with the highest priority to flush the
                    # last check-in. This task will be processed by the writer
                    # thread immediately.
                    self._schedule(TimedTask(when=0, flag_name=flag_name))

                # As there may be no flag names at all, the queue may be
                # untouched. In this case, we need to notify the writer thread
                # explicitly.
                with self.refresh_condition:
                    self.refresh_condition.notify()

                # If `block` is set, we wait until the writer thread finishes
                # flushing the last check-in.
                if block:
                    logger.debug("blocking until the writer finishes...")
                    self.flush_condition.wait()
                    logger.debug("done")

    def _check_in(
        self,
        *,
        expect_next_in: int = 0,
        flag_name: str = "check_in",
        flush: bool = False,
        block: bool = False,
    ) -> None:
        """
        Check-in and optionally mark an expiry date by setting `expect_next_in`.

        If `block` is True, then this will block until the check-in is flushed.
        """
        if block:
            flush = True

        if self.stop_signal.is_set():
            raise RuntimeError("check-in is not allowed after stopping the writer thread")

        with self.reporter_lock:
            self._increment(
                expect_next_in=expect_next_in,
                flag_name=flag_name,
            )
            if flush:
                self.flush(block=block)

    def _report_successful_finish(
        self,
        *,
        flush: bool = True,
        block: bool = True,
    ) -> None:
        """
        Report a successful end of the Run.

        By default this will block until the check-in is flushed.
        """
        return self._check_in(flush=flush, block=block, flag_name="finished")

    # The instance method `report_progress` and `report_successful_finish` is patched into
    # the instance in `__post_init__`
    # so the same name is used for both the instance method and the class method.
    @classmethod
    def report_progress(
        cls,
        *,
        expect_next_in: int = 0,
        flush: bool = False,
        block: bool = False,
    ) -> None:
        """
        Check-in and optionally mark an expiry date by setting `expect_next_in`.

        If `block` is True, then this will block until the check-in is flushed.

        Note: This is a classmethod and designed to be called like:
        `RunStatusReporter.report_progress(expect_next_in=10)`
        * If no instance is available, this will log a warning and return.
        * If multiple instances are available, this will raise an error.
        """
        default_instance = cls.default()
        if default_instance is None:
            return
        default_instance._check_in(expect_next_in=expect_next_in, flush=flush, block=block)

    @classmethod
    def report_successful_finish(
        cls,
        *,
        flush: bool = True,
        block: bool = True,
    ) -> None:
        """
        Report a successful end of the Run.

        By default this will block until the check-in is flushed.

        Note: This is a classmethod and designed to be called like:
        `RunStatusReporter.report_successful_finish()`
        * If no instance is available, this will log a warning and return.
        * If multiple instances are available, this will raise an error.
        """
        default_instance = cls.default()
        if default_instance is None:
            return
        default_instance._report_successful_finish(flush=flush, block=block)


REPORT_INTERVAL = 5  # 5 seconds


class ScheduledStatusReporter(object):
    def __init__(self,
                 status_reporter: RunStatusReporter,
                 flag: str = 'progress',
                 interval: int = REPORT_INTERVAL
                 ):
        self.status_reporter = status_reporter
        self.flag = flag
        self.report_interval = interval
        self.throttle = 30
        self.thread = threading.Thread(target=self._run, daemon=True)
        self.stop_signal = threading.Event()
        self.thread.start()

    def _run(self):
        self.status_reporter.check_in(expect_next_in=self.throttle, flag_name=self.flag, block=True)

        while True:
            if self.stop_signal.wait(timeout=self.report_interval):
                # report last heartbeat
                self.status_reporter.check_in(expect_next_in=1, flag_name=self.flag, block=True)
                break
            else:
                self.status_reporter.check_in(expect_next_in=self.throttle, flag_name=self.flag)

    def stop(self):
        self.stop_signal.set()
        self.thread.join()



import asyncio
import time

from typing import Iterable, Iterator, List, Tuple, Union, Optional
from typing import TYPE_CHECKING

from aim.web.api.runs.utils import (
    ASYNC_SLEEP_INTERVAL,
    PROGRESS_KEY_SUFFIX,
    IndexRange,
    collect_streamable_data,
    get_run_props,
    get_run_params,
)
from aim.web.configs import AIM_PROGRESS_REPORT_INTERVAL
from aim.sdk.uri_service import URIService, generate_resource_path
from aim.sdk.sequence_collection import SequenceCollection
from aim.sdk.sequence import Sequence

from aim.storage.treeutils import encode_tree
from aim.storage.context import Context

if TYPE_CHECKING:
    from aim.sdk.repo import Repo
    from aim.sdk.run import Run


def get_blobs_batch(uri_batch: List[str], repo: 'Repo') -> Iterator[bytes]:
    uri_service = URIService(repo=repo)
    batch_iterator = uri_service.request_batch(uri_batch=uri_batch)
    for it in batch_iterator:
        yield collect_streamable_data(encode_tree(it))


class CustomObjectApi:
    def __init__(self, seq_type: str, resolve_blobs: bool):
        seq_cls = Sequence.registry.get(seq_type, None)
        if seq_cls is None:
            raise ValueError(f'\'{self.seq_type}\' is not a valid sequence type.')
        self.seq_type = seq_cls.sequence_name()
        self.resolve_blobs = resolve_blobs
        self.use_list = seq_cls.collections_allowed
        self.dump_fn = None

        self.traces: SequenceCollection = None
        self.requested_traces: list = []
        self.run: Run = None
        self.trace_cache: dict = {}

        self.record_range = None
        self.record_density = None
        self.total_record_range = None
        self.index_range = None
        self.index_density = None
        self.index_slice = None
        self.total_index_range = None

        if self.use_list:
            self._value_retriever = self._record_collection_retriever
        else:
            self._value_retriever = self._record_to_encodable

    def set_dump_data_fn(self, dump_fn: callable):
        self.dump_fn = dump_fn

    def set_trace_collection(self, traces: SequenceCollection):
        self.traces = traces

    def set_requested_traces(self, run: 'Run', requested_traces):
        self.run = run
        for requested_trace in requested_traces:
            trace_name = requested_trace.name
            context = Context(requested_trace.context)
            trace = run._get_sequence(self.seq_type, trace_name, context)
            if not trace:
                continue
            self.requested_traces.append(trace)

    def set_ranges(self,
                   record_range: IndexRange,
                   record_density: int,
                   index_range: Optional[IndexRange] = None,
                   index_density: Optional[int] = None,
                   record_step: Optional[int] = None):
        self.record_range = record_range
        self.record_density = record_density

        if self.use_list:
            assert index_range is not None
            assert index_density is not None
            self.index_range = index_range
            self.index_density = index_density
            self.total_record_range, self.total_index_range = self._calculate_ranges()
        else:
            self.total_record_range = self._calculate_ranges()

        if record_step is not None:
            if record_step < 0:
                record_step += self.total_record_range.stop
            assert record_step < self.total_record_range.stop

            self.record_range = IndexRange(record_step, record_step + 1)

        self._adjust_ranges()
        if self.use_list:
            step = (self.index_range.stop - self.index_range.start) // index_density or 1
            self.index_slice = slice(self.index_range.start, self.index_range.stop, step)

    def get_total_record_range(self):
        return self._calculate_ranges()

    async def search_result_streamer(self, skip_system: bool, report_progress: bool):
        def _pack_run_data(run_: 'Run', traces_: list):
            ranges = {
                'record_range_used': self.record_range,
                'record_range_total': self.total_record_range
            }
            if self.use_list:
                ranges['index_range_used'] = self.index_range
                ranges['index_range_total'] = self.total_index_range
            run_dict = {
                run_.hash: {
                    'ranges': ranges,
                    'params': get_run_params(run_, skip_system=skip_system),
                    'traces': traces_,
                    'props': get_run_props(run_)
                }
            }
            return collect_streamable_data(encode_tree(run_dict))
        try:
            last_reported_progress_time = time.time()
            run_info = None
            progress_reports_sent = 0
            for run_info in self.trace_cache.values():
                await asyncio.sleep(ASYNC_SLEEP_INTERVAL)
                if report_progress and time.time() - last_reported_progress_time > AIM_PROGRESS_REPORT_INTERVAL:
                    yield collect_streamable_data(encode_tree(
                        {f'progress_{progress_reports_sent}_{PROGRESS_KEY_SUFFIX}': run_info['progress']}
                    ))
                    yield collect_streamable_data(encode_tree({f'progress_{progress_reports_sent}':
                                                               run_info['progress']}))
                    progress_reports_sent += 1
                    last_reported_progress_time = time.time()
                if run_info.get('traces') and run_info.get('run'):
                    traces_list = []
                    for trace in run_info['traces']:
                        traces_list.append(self._get_trace_info(trace, True, True))
                    yield _pack_run_data(run_info['run'], traces_list)
                    if report_progress:
                        yield collect_streamable_data(encode_tree({f'progress_{progress_reports_sent}':
                                                                   run_info['progress']}))
                        progress_reports_sent += 1
                        last_reported_progress_time = time.time()

            if report_progress and run_info:
                yield collect_streamable_data(encode_tree({f'progress_{progress_reports_sent}':
                                                           run_info['progress']}))
        except asyncio.CancelledError:
            pass

    async def requested_traces_streamer(self) -> List[dict]:
        try:
            for run_info in self.trace_cache.values():
                await asyncio.sleep(ASYNC_SLEEP_INTERVAL)
                for trace in run_info['traces']:
                    trace_dict = self._get_trace_info(trace, False, False)
                    trace_dict['record_range_used'] = self.record_range
                    trace_dict['record_range_total'] = self.total_record_range
                    if self.use_list:
                        trace_dict['index_range'] = self.index_range
                        trace_dict['index_range_total'] = self.total_index_range
                    yield collect_streamable_data(encode_tree(trace_dict))
        except asyncio.CancelledError:
            pass

    def _get_trace_info(self, trace: Sequence, include_epochs: bool, include_timestamps: bool) -> dict:
        steps = []
        values = []

        steps_vals = trace.data.view('val').\
            range(self.record_range.start, self.record_range.stop).\
            sample(self.record_density)

        for step, (val,) in steps_vals:
            steps.append(step)
            values.append(self._value_retriever(step, val, trace))

        result = {
            'name': trace.name,
            'context': trace.context.to_dict(),
            'values': values,
            'iters': steps
        }
        if include_epochs:
            result['epochs'] = trace.data.view('epoch').\
                range(self.record_range.start, self.record_range.stop).\
                sample(self.record_density).values_list()
        if include_timestamps:
            result['timestamps'] = trace.data.view('time').\
                range(self.record_range.start, self.record_range.stop).\
                sample(self.record_density).values_list()
        return result

    def _record_to_encodable(self, step: int, record, trace: Sequence) -> dict:
        rec_json = record.json()
        if self.resolve_blobs:
            rec_json['data'] = self.dump_fn(record)
        else:
            # TODO make path generation more abstract
            data_path = generate_resource_path(trace.values.tree.container, (step, 'data'))
            rec_json['blob_uri'] = URIService.generate_uri(trace.run.repo, trace.run.hash, 'seqs', data_path)
        return rec_json

    def _record_collection_to_encodable(self, step: int, record_collection: Iterable, trace: Sequence) -> List[dict]:
        assert self.use_list, f'Sequence \'{self.seq_type}\' does not support record collections.'
        result = []
        for idx, record in record_collection:
            rec_json = record.json()
            rec_json['index'] = idx
            if self.resolve_blobs:
                rec_json['data'] = self.dump_fn(record)
            else:
                # TODO make path generation more abstract
                data_path = generate_resource_path(trace.values.tree.container, (step, idx, 'data'))
                rec_json['blob_uri'] = URIService.generate_uri(trace.run.repo, trace.run.hash, 'seqs', data_path)
            result.append(rec_json)
        return result

    def _record_collection_retriever(self, step, val, trace):
        assert self.use_list
        if isinstance(val, list):
            sliced_val = self.record_collection_slice(val, self.index_slice)
            return self._record_collection_to_encodable(step, sliced_val, trace)
        elif self.index_range.start == 0:
            res = self._record_to_encodable(step, val, trace)
            res['index'] = 0
            return [res]
        else:
            return []

    @staticmethod
    def record_collection_slice(values, _slice: slice) -> Iterable:
        yield from zip(range(_slice.start, _slice.stop, _slice.step), values[_slice])

    def _foreach_trace(self, callback: callable):
        if self.traces:
            for run_trace_collection, progress in self.traces.iter_runs():
                if not run_trace_collection:
                    continue
                run = run_trace_collection.run
                self.trace_cache[run.hash] = {'progress': progress}
                run_traces = []
                for trace in run_trace_collection.iter():
                    run_traces.append(trace)
                    callback(trace)
                if run_traces:
                    self.trace_cache[run.hash].update({
                        'run': run,
                        'traces': run_traces
                    })
        elif self.requested_traces:
            for trace in self.requested_traces:
                callback(trace)
            assert self.run is not None
            self.trace_cache[self.run.hash] = {
                'run': self.run,
                'traces': self.requested_traces
            }

    def _calculate_ranges(self) -> Union[IndexRange, Tuple[IndexRange, IndexRange]]:
        rec_start = None
        rec_stop = -1
        idx_start = 0  # record inner indexing is always sequential
        idx_stop = -1

        if self.use_list:
            def _update_ranges(trace: Sequence):
                nonlocal rec_start, rec_stop, idx_stop
                rec_start = min(trace.first_step(), rec_start) if rec_start else trace.first_step()
                rec_stop = max(trace.last_step(), rec_stop)
                idx_stop = max(trace.record_length() or 1, idx_stop)
        else:
            def _update_ranges(trace: Sequence):
                nonlocal rec_start, rec_stop, idx_stop
                rec_start = min(trace.first_step(), rec_start) if rec_start else trace.first_step()
                rec_stop = max(trace.last_step(), rec_stop)

        self._foreach_trace(_update_ranges)

        if self.use_list:
            return IndexRange(rec_start, rec_stop + 1), IndexRange(idx_start, idx_stop)
        else:
            return IndexRange(rec_start, rec_stop + 1)

    def _adjust_ranges(self):
        start = self.record_range.start if self.record_range.start is not None else self.total_record_range.start
        stop = self.record_range.stop if self.record_range.stop is not None else self.total_record_range.stop
        self.record_range = IndexRange(start, stop)
        if self.use_list:
            start = self.index_range.start if self.index_range.start is not None else self.total_index_range.start
            stop = self.index_range.stop if self.index_range.stop is not None else self.total_index_range.stop
            self.index_range = IndexRange(start, stop)



"""
This code is taken from https://github.com/sunhailin-Leo/fastapi_profiler
Credits to the author of the repo.
"""

import json
import os.path
import time
from typing import Optional
from logging import getLogger

from starlette.routing import Router
from starlette.requests import Request
from starlette.types import ASGIApp, Message, Receive, Scope, Send

logger = getLogger("profiler")


class PyInstrumentProfilerMiddleware:

    def __init__(
            self, app: ASGIApp,
            *,
            server_app: Optional[Router] = None,
            profiler_interval: float = 0.0001,
            repo_path=None,
            **profiler_kwargs
    ):

        try:
            from pyinstrument import Profiler, __version__

            if tuple(map(int, __version__.split('.'))) < (3,):
                raise ImportError
        except ImportError:
            raise RuntimeError(
                'This contrib module requires \'pyinstrument\' to be installed. '
                'Please install it with command: \n pip install \'pyinstrument>=3.0.0\''
            )

        self.profiler = Profiler
        self.app = app

        self._server_app = server_app
        self._profiler_kwargs: dict = profiler_kwargs

        self._profiler_interval = profiler_interval
        self._profiler_timestamp = str(time.time())
        self._profiler_log_path = os.path.join(repo_path, "profiler", self._profiler_timestamp)
        os.makedirs(self._profiler_log_path, exist_ok=True)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = Request(scope, receive=receive)

        request_time = time.time()
        profiler = self.profiler(interval=self._profiler_interval)
        try:
            profiler.start()
        except: # noqa
            skip_profiling = True
        else:
            skip_profiling = False

        # Default status code used when the application does not return a valid response
        # or an unhandled exception occurs.
        status_code = 500

        async def wrapped_send(message: Message) -> None:
            if message['type'] == 'http.response.start':
                nonlocal status_code
                status_code = message['status']
            await send(message)

        try:
            await self.app(scope, receive, wrapped_send)
        finally:
            if skip_profiling:
                return

            profiler.stop()

            method = request.method
            path = request.url.path
            params = dict(request.query_params)

            file_name = (
                '{timestamp}_{method}_{path}'.format(
                    timestamp=request_time,
                    method=method.lower(),
                    path='_'.join(path.strip('/').split('/')).lower()
                )
            )
            request_data = json.dumps({
                'path': path,
                'method': method,
                'params': params
            }, separators=(',', ':'))

            # inject request data
            html_output = profiler.output_html(**self._profiler_kwargs)
            body_tag = '<body>'
            body_tag_idx_end = html_output.find(body_tag) + len(body_tag)
            html_output = (
                f'{html_output[:body_tag_idx_end]}'
                f'<pre><code>{request_data}</code></pre>'
                f'{html_output[body_tag_idx_end:]}'
            )

            with open(os.path.join(self._profiler_log_path, f'{file_name}.html'), 'w') as fp:
                fp.write(html_output)



import logging
from typing import Dict, Optional, Union
from typing import TYPE_CHECKING

from aim.storage.hashing import hash_auto
from aim.storage.treeview import TreeView
from aim.sdk.utils import generate_run_hash
from aim.sdk.repo_utils import get_repo
from aim.sdk.errors import MissingRunError
from aim.sdk.tracker import STEP_HASH_FUNCTIONS

if TYPE_CHECKING:
    from aim.sdk.repo import Repo

logger = logging.getLogger(__name__)


class BaseRun:
    def __new__(cls, *args, **kwargs):
        # prevent BaseRun from being instantiated directly
        if cls is BaseRun:
            raise TypeError(f'Only children of \'{cls.__name__}\' may be instantiated.')
        return object.__new__(cls)

    def __init__(self, run_hash: Optional[str] = None,
                 repo: Optional[Union[str, 'Repo']] = None,
                 read_only: bool = False,
                 force_resume: bool = False):
        self._hash = None
        self._lock = None

        self.read_only = read_only
        self.repo = get_repo(repo)
        if self.read_only:
            assert run_hash is not None
            self.hash = run_hash
        else:
            if run_hash is None:
                self.hash = generate_run_hash()
            elif self.repo.run_exists(run_hash):
                self.hash = run_hash
            else:
                raise MissingRunError(f'Cannot find Run {run_hash} in aim Repo {self.repo.path}.')
            self._lock = self.repo.request_run_lock(self.hash)
            self._lock.lock(force=force_resume)

        self.meta_tree: TreeView = self.repo.request_tree(
            'meta', self.hash, read_only=read_only, from_union=True
        ).subtree('meta')
        self.meta_run_tree: TreeView = self.meta_tree.subtree('chunks').subtree(self.hash)

        self._series_run_trees: Dict[int, TreeView] = None

    def __hash__(self) -> int:
        if self._hash is None:
            self._hash = self._calc_hash()
        return self._hash

    def __repr__(self) -> str:
        return f'<Run#{hash(self)} name={self.hash} repo={self.repo}>'

    @property
    def series_run_trees(self) -> Dict[int, TreeView]:
        if self._series_run_trees is None:
            series_tree = self.repo.request_tree(
                'seqs', self.hash, read_only=self.read_only
            ).subtree('seqs')
            self._series_run_trees = {}
            for version in STEP_HASH_FUNCTIONS.keys():
                if version == 1:
                    self._series_run_trees[version] = series_tree.subtree(('chunks', self.hash))
                else:
                    self._series_run_trees[version] = series_tree.subtree((f'v{version}', 'chunks', self.hash))
        return self._series_run_trees

    def check_metrics_version(self) -> bool:
        metric_dtypes = ('float', 'float64', 'int')
        traces_tree = self.meta_run_tree.get('traces', {})

        v1_metric_found = False
        for ctx_metadata in traces_tree.values():
            for seq_metadata in ctx_metadata.values():
                if seq_metadata.get('dtype', 'float') in metric_dtypes:
                    if seq_metadata.get('version', 1) == 1:
                        v1_metric_found = True
                        break
        return v1_metric_found

    def update_metrics(self):
        metric_dtypes = ('float', 'float64', 'int')
        series_meta_tree = self.meta_run_tree.subtree('traces')

        for ctx_id, ctx_traces in series_meta_tree.items():
            for name, trace_info in ctx_traces.items():
                if (trace_info.get('dtype', 'float') in metric_dtypes) and (trace_info.get('version', 1) == 1):
                    series = self.series_run_trees[1].subtree((ctx_id, name))
                    new_series = self.series_run_trees[2].subtree((ctx_id, name))
                    step_view = new_series.array('step', dtype='int64').allocate()
                    val_view = new_series.array('val').allocate()
                    epoch_view = new_series.array('epoch', dtype='int64').allocate()
                    time_view = new_series.array('time', dtype='int64').allocate()
                    for (step, val), (_, epoch), (_, timestamp) in zip(
                            series.subtree('val').items(),
                            series.subtree('epoch').items(),
                            series.subtree('time').items()):
                        step_hash = hash_auto(step)
                        step_view[step_hash] = step
                        val_view[step_hash] = val
                        epoch_view[step_hash] = epoch
                        time_view[step_hash] = timestamp
                    self.meta_run_tree['traces', ctx_id, name, 'version'] = 2
                    del self.series_run_trees[1][(ctx_id, name)]

    def _calc_hash(self) -> int:
        # TODO maybe take read_only flag into account?
        return hash_auto((self.hash, hash(self.repo)))



from typing import Optional, Dict, List

from fastapi import HTTPException, Header
from pydantic import BaseModel
from starlette.responses import StreamingResponse

from aim import Images, Texts, Distributions, Audios, Figures
from aim.sdk.sequence import Sequence
from aim.sdk.sequence_collection import QuerySequenceCollection
from aim.sdk.types import QueryReportMode
from aim.web.api.runs.pydantic_models import (
    RunTracesBatchApiIn,
    URIBatchIn,
    QuerySyntaxErrorOut,
    ImageList,
    TextList,
    AudioList,
    DistributionInfo,
    FigureInfo,
    ObjectSearchRunView,
    ObjectSequenceBaseView,
)
from aim.web.api.runs.utils import (
    checked_query,
    checked_range,
    get_project_repo,
    numpy_to_encodable,
    get_run_or_404
)
from aim.web.api.runs.object_api_utils import CustomObjectApi, get_blobs_batch


class CustomObjectApiConfig:
    sequence_type: type = Sequence
    resolve_blobs: bool = False
    dump_record_fn: callable = lambda x: x.data  # noqa E731
    model: type = BaseModel

    @staticmethod
    def check_density(density):
        if density <= 0:
            raise HTTPException(status_code=400, detail='Density must be greater than 0.')

    @classmethod
    def register_endpoints(cls, router):
        assert issubclass(cls.sequence_type, Sequence)
        seq_name = cls.sequence_type.sequence_name()

        # search API
        search_endpoint = f'/search/{seq_name}/'

        @router.get(search_endpoint, response_model=Dict[str, ObjectSearchRunView],
                    responses={400: {'model': QuerySyntaxErrorOut}})
        async def search_api(q: Optional[str] = '',
                             skip_system: Optional[bool] = True,
                             record_range: Optional[str] = '', record_density: Optional[int] = 50,
                             index_range: Optional[str] = '', index_density: Optional[int] = 5,
                             report_progress: Optional[bool] = True,
                             x_timezone_offset: int = Header(default=0),):
            # search Sequence API
            repo = get_project_repo()
            query = checked_query(q)
            record_range = checked_range(record_range)
            index_range = checked_range(index_range)
            CustomObjectApiConfig.check_density(record_density)
            CustomObjectApiConfig.check_density(index_density)

            # TODO [MV, AT]: move to `repo.py` when `SELECT` statements are introduced
            repo._prepare_runs_cache()
            query_iterator = QuerySequenceCollection(repo=repo,
                                                     seq_cls=cls.sequence_type,
                                                     query=query,
                                                     report_mode=QueryReportMode.PROGRESS_TUPLE,
                                                     timezone_offset=x_timezone_offset)

            api = CustomObjectApi(seq_name, resolve_blobs=cls.resolve_blobs)
            api.set_dump_data_fn(cls.dump_record_fn)
            api.set_trace_collection(query_iterator)
            api.set_ranges(record_range, record_density, index_range, index_density)
            streamer = api.search_result_streamer(skip_system, report_progress)
            return StreamingResponse(streamer)

        # run sequence batch API
        sequence_batch_endpoint = f'/{{run_id}}/{seq_name}/get-batch/'

        @router.post(sequence_batch_endpoint, response_model=List[ObjectSequenceBaseView])
        async def sequence_batch_api(run_id: str,
                                     requested_traces: RunTracesBatchApiIn,
                                     record_range: Optional[str] = '', record_density: Optional[int] = 50,
                                     index_range: Optional[str] = '', index_density: Optional[int] = 5):
            # get Sequence batch API
            record_range = checked_range(record_range)
            index_range = checked_range(index_range)
            CustomObjectApiConfig.check_density(record_density)
            CustomObjectApiConfig.check_density(index_density)

            run = get_run_or_404(run_id)

            api = CustomObjectApi(seq_name, resolve_blobs=cls.resolve_blobs)
            api.set_dump_data_fn(cls.dump_record_fn)
            api.set_requested_traces(run, requested_traces)
            api.set_ranges(record_range, record_density, index_range, index_density)
            traces_streamer = api.requested_traces_streamer()
            return StreamingResponse(traces_streamer)

        if not cls.resolve_blobs:
            # get BLOB batch API
            uri_batch_endpoint = f'/{seq_name}/get-batch'

            @router.post(uri_batch_endpoint)
            def blobs_batch_api(uri_batch: URIBatchIn):
                return StreamingResponse(get_blobs_batch(uri_batch, get_project_repo()))

        # run sequence batch API
        step_of_sequence_endpoint = f'/{{run_id}}/{seq_name}/get-step/'

        @router.post(step_of_sequence_endpoint, response_model=List[ObjectSequenceBaseView])
        async def step_of_sequence(run_id: str,
                                   requested_traces: RunTracesBatchApiIn,
                                   index_range: Optional[str] = '', index_density: Optional[int] = 5,
                                   record_step: int = -1):
            # get last step by default

            index_range = checked_range(index_range)
            CustomObjectApiConfig.check_density(index_density)

            run = get_run_or_404(run_id)

            api = CustomObjectApi(seq_name, resolve_blobs=cls.resolve_blobs)
            api.set_dump_data_fn(cls.dump_record_fn)
            api.set_requested_traces(run, requested_traces)

            api.set_ranges(None, 1, index_range, index_density, record_step)
            traces_streamer = api.requested_traces_streamer()
            return StreamingResponse(traces_streamer)


class ImageApiConfig(CustomObjectApiConfig):
    sequence_type = Images
    resolve_blobs = False
    model = ImageList


class TextApiConfig(CustomObjectApiConfig):
    sequence_type = Texts
    resolve_blobs = True
    model = TextList


class DistributionApiConfig(CustomObjectApiConfig):
    sequence_type = Distributions
    resolve_blobs = True
    dump_record_fn = lambda x: numpy_to_encodable(x.weights)  # noqa E731
    model = DistributionInfo


class AudioApiConfig(CustomObjectApiConfig):
    sequence_type = Audios
    resolve_blobs = False
    model = AudioList


class FigureApiConfig(CustomObjectApiConfig):
    sequence_type = Figures
    resolve_blobs = True
    dump_record_fn = lambda x: x.data  # noqa E731
    model = FigureInfo
